#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from argparse import ArgumentParser

from logging import basicConfig as logging_basicConfig
from logging import getLevelName as logging_getLevelName
# from logging import debug as logging_debug
#from logging import info as logging_info

import tkinter as tk
from tkinter import ttk

from MethodicConfigurator.common_arguments import add_common_arguments_and_parse

from MethodicConfigurator.backend_filesystem import LocalFilesystem

from MethodicConfigurator.battery_cell_voltages import BatteryCell

from MethodicConfigurator.frontend_tkinter_component_editor_base import ComponentEditorWindowBase

#from MethodicConfigurator.frontend_tkinter_base import show_tooltip
from MethodicConfigurator.frontend_tkinter_base import show_error_message

from MethodicConfigurator.version import VERSION


def argument_parser():
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.
    """
    parser = ArgumentParser(description='A GUI for editing JSON files that contain vehicle component configurations. '
                            'Not to be used directly, but through the main ArduPilot methodic configurator script.')
    parser = LocalFilesystem.add_argparse_arguments(parser)
    parser = ComponentEditorWindow.add_argparse_arguments(parser)
    return add_common_arguments_and_parse(parser)


class VoltageTooLowError(Exception):
    """Raised when the voltage is below the minimum limit."""


class VoltageTooHighError(Exception):
    """Raised when the voltage is above the maximum limit."""

class ComponentEditorWindow(ComponentEditorWindowBase):
    """
    This class validates the user input and handles user interactions
    for editing component configurations in the ArduPilot Methodic Configurator.
    """
    def __init__(self, version, local_filesystem: LocalFilesystem=None):
        ComponentEditorWindowBase.__init__(self, version, local_filesystem)

    def update_json_data(self):
        super().update_json_data()
        if 'Components' not in self.data:
            self.data['Components'] = {}
        if 'Battery' not in self.data['Components']:
            self.data['Components']['Battery'] = {}
        if 'Specifications' not in self.data['Components']['Battery']:
            self.data['Components']['Battery']['Specifications'] = {}
        if 'Chemistry' not in self.data['Components']['Battery']['Specifications']:
            self.data['Components']['Battery']['Specifications']['Chemistry'] = "Lipo"
        if 'Capacity mAh' not in self.data['Components']['Battery']['Specifications']:
            self.data['Components']['Battery']['Specifications']['Capacity mAh'] = 0
        if 'Frame' not in self.data['Components']:
            self.data['Components']['Frame'] = {}
        if 'Specifications' not in self.data['Components']['Frame']:
            self.data['Components']['Frame']['Specifications'] = {}
        if 'TOW min Kg' not in self.data['Components']['Frame']['Specifications']:
            self.data['Components']['Frame']['Specifications']['TOW min Kg'] = 1
        if 'TOW max Kg' not in self.data['Components']['Frame']['Specifications']:
            self.data['Components']['Frame']['Specifications']['TOW max Kg'] = 1

    def set_vehicle_type_and_version(self, vehicle_type: str, version: str):
        self._set_component_value_and_update_ui(('Flight Controller', 'Firmware', 'Type'), vehicle_type)
        if version:
            self._set_component_value_and_update_ui(('Flight Controller', 'Firmware', 'Version'), version)

    def set_fc_manufacturer(self, manufacturer: str):
        if manufacturer and manufacturer!= "Unknown" and manufacturer!= "ArduPilot":
            self._set_component_value_and_update_ui(('Flight Controller', 'Product', 'Manufacturer'), manufacturer)

    def set_fc_model(self, model: str):
        if model and model!= "Unknown" and model!= "MAVLink":
            self._set_component_value_and_update_ui(('Flight Controller', 'Product', 'Model'), model)

    @staticmethod
    def reverse_key_search(doc: dict, param_name: str, values: list) -> list:
        return [key for key, value in doc[param_name]["values"].items() if value in values]

    def set_values_from_fc_parameters(self, fc_parameters: dict, doc: dict):
        serial_ports = ["SERIAL1", "SERIAL2", "SERIAL3", "SERIAL4", "SERIAL5", "SERIAL6", "SERIAL7", "SERIAL8"]
        #can_ports = ["CAN1", "CAN2"]
        #i2c_ports = ["I2C1", "I2C2", "I2C3", "I2C4"]

        rc_receiver_protocols = self.reverse_key_search(doc, "SERIAL1_PROTOCOL", ["RC Input"])
        telemetry_protocols = self.reverse_key_search(doc, "SERIAL1_PROTOCOL",
                                                      ["MAVLink1", "MAVLink2", "MAVLink High Latency"])
        gnss_protocols = self.reverse_key_search(doc, "SERIAL1_PROTOCOL", ["GPS"])
        esc_protocols = self.reverse_key_search(doc, "SERIAL1_PROTOCOL", ["ESC Telemetry", "FETtecOneWire", "CoDevESC"])
        for serial in serial_ports:
            if serial + "_PROTOCOL" in fc_parameters:
                if fc_parameters[serial + "_PROTOCOL"] in rc_receiver_protocols:
                    self.data['Components']['RC Receiver']['FC Connection']['Type'] = serial
                    #self.data['Components']['RC Receiver']['FC Connection']['Protocol'] = \
                    # doc['RC_PROTOCOLS']['values'][fc_parameters['RC_PROTOCOLS']]
                elif fc_parameters[serial + "_PROTOCOL"] in telemetry_protocols:
                    self.data['Components']['Telemetry']['FC Connection']['Type'] = serial
                    self.data['Components']['Telemetry']['FC Connection']['Protocol'] = \
                        doc[serial + "_PROTOCOL"]['values'][str(fc_parameters[serial + "_PROTOCOL"]).rstrip('0').rstrip('.')]
                elif fc_parameters[serial + "_PROTOCOL"] in gnss_protocols:
                    self.data['Components']['GNSS Receiver']['FC Connection']['Type'] = serial
                    self.data['Components']['GNSS Receiver']['FC Connection']['Protocol'] = \
                        doc['GPS_TYPE']['values'][str(fc_parameters['GPS_TYPE']).rstrip('0').rstrip('.')]
                elif fc_parameters[serial + "_PROTOCOL"] in esc_protocols:
                    self.data['Components']['ESC']['FC Connection']['Type'] = serial
                    self.data['Components']['ESC']['FC Connection']['Protocol'] = \
                        doc['MOT_PWM_TYPE']['values'][str(fc_parameters['MOT_PWM_TYPE']).rstrip('0').rstrip('.')]
        if "BATT_MONITOR" in fc_parameters:
            analog = [key for key, value in doc["BATT_MONITOR"]["values"].items() \
                      if value in ['Analog Voltage Only', 'Analog Voltage and Current']]
            if fc_parameters["BATT_MONITOR"] in analog:
                self.data['Components']['Battery Monitor']['FC Connection']['Type'] = "Analog"
            self.data['Components']['Battery Monitor']['FC Connection']['Protocol'] = \
                doc['BATT_MONITOR']['values'][str(fc_parameters["BATT_MONITOR"]).rstrip('0').rstrip('.')]

    def add_entry_or_combobox(self, value, entry_frame, path):
        serial_ports = ["SERIAL1", "SERIAL2", "SERIAL3", "SERIAL4", "SERIAL5", "SERIAL6", "SERIAL7", "SERIAL8"]
        can_ports = ["CAN1", "CAN2"]
        i2c_ports = ["I2C1", "I2C2", "I2C3", "I2C4"]

        combobox_config = {
            ('Flight Controller', 'Firmware', 'Type'): {
                "values": LocalFilesystem.supported_vehicles(),
            },
            ('RC Receiver', 'FC Connection', 'Type'): {
                "values": ["RCin/SBUS"] + serial_ports + can_ports,
            },
            ('RC Receiver', 'FC Connection', 'Protocol'): {
                "values": ["All", "PPM", "IBUS", "SBUS", "SBUS_NI", "DSM", "SUMD", "SRXL", "SRXL2",
                           "CRSF", "ST24", "FPORT", "FPORT2", "FastSBUS", "DroneCAN", "Ghost", "MAVRadio"],
            },
            ('Telemetry', 'FC Connection', 'Type'): {
                "values": serial_ports + can_ports,
            },
            ('Telemetry', 'FC Connection', 'Protocol'): {
                # TODO get this list from SERIAL1_PROTOCOL pdef metadata pylint: disable=fixme
                "values": ["MAVLink1", "MAVLink2", "MAVLink High Latency"],
            },
            ('Battery Monitor', 'FC Connection', 'Type'): {
                "values": ['Analog'] + i2c_ports + serial_ports + can_ports,
            },
            ('Battery Monitor', 'FC Connection', 'Protocol'): {
                # TODO get this list from BATT_MONITOR pdef metadata pylint: disable=fixme
                "values": ['Analog Voltage Only', 'Analog Voltage and Current', 'Solo', 'Bebop', 'SMBus-Generic',
                           'DroneCAN-BatteryInfo', 'ESC', 'Sum Of Selected Monitors', 'FuelFlow', 'FuelLevelPWM',
                           'SMBUS-SUI3', 'SMBUS-SUI6', 'NeoDesign', 'SMBus-Maxell', 'Generator-Elec', 'Generator-Fuel',
                           'Rotoye', 'MPPT', 'INA2XX', 'LTC2946', 'Torqeedo', 'FuelLevelAnalog',
                           'Synthetic Current and Analog Voltage', 'INA239_SPI', 'EFI', 'AD7091R5', 'Scripting'],
            },
            ('ESC', 'FC Connection', 'Type'): {
                "values": ['Main Out', 'AIO'] + serial_ports + can_ports,
            },
            ('ESC', 'FC Connection', 'Protocol'): {
                # TODO get this list from MOT_PWM_TYPE pdef metadata pylint: disable=fixme
                "values": ['Normal', 'OneShot', 'OneShot125', 'Brushed', 'DShot150', 'DShot300', 'DShot600',
                           'DShot1200', 'PWMRange', 'PWMAngle'],
            },
            ('GNSS receiver', 'FC Connection', 'Type'): {
                "values": serial_ports + can_ports,
            },
            ('GNSS receiver', 'FC Connection', 'Protocol'): {
                # TODO get this list from GPS_TYPE pdef metadata pylint: disable=fixme
                "values": ['Auto', 'uBlox', 'NMEA', 'SiRF', 'HIL', 'SwiftNav', 'DroneCAN', 'SBF', 'GSOF', 'ERB',
                           'MAV', 'NOVA', 'HemisphereNMEA', 'uBlox-MovingBaseline-Base', 'uBlox-MovingBaseline-Rover',
                           'MSP', 'AllyStar', 'ExternalAHRS', 'Unicore', 'DroneCAN-MovingBaseline-Base',
                           'DroneCAN-MovingBaseline-Rover', 'UnicoreNMEA', 'UnicoreMovingBaselineNMEA', 'SBF-DualAntenna'],
            },
            ('Battery', 'Specifications', 'Chemistry'): {
                "values": BatteryCell.chemistries(),
            },
        }
        config = combobox_config.get(path)
        if config:
            cb = ttk.Combobox(entry_frame, values=config["values"])
            cb.bind("<FocusOut>", lambda event, path=path: self.validate_combobox(event, path))
            cb.bind("<KeyRelease>", lambda event, path=path: self.validate_combobox(event, path))
            cb.set(value)
            return cb

        entry = ttk.Entry(entry_frame)
        validate_function = self.get_validate_function(entry, path)
        if validate_function:
            entry.bind("<FocusOut>", validate_function)
            entry.bind("<KeyRelease>", validate_function)
        entry.insert(0, str(value))
        return entry

    def get_validate_function(self, entry, path):
        validate_functions = {
            ('Frame', 'Specifications', 'TOW min Kg'): lambda event, entry=entry, path=path:
                self.validate_entry_limits(event, entry, float, (0.01, 600), "Takeoff Weight", path),

            ('Frame', 'Specifications', 'TOW max Kg'): lambda event, entry=entry, path=path:
                self.validate_entry_limits(event, entry, float, (0.01, 600), "Takeoff Weight", path),

            ('Battery', 'Specifications', 'Volt per cell max'): lambda event, entry=entry, path=path:
                self.validate_cell_voltage(event, entry, path),

            ('Battery', 'Specifications', 'Volt per cell low'): lambda event, entry=entry, path=path:
                self.validate_cell_voltage(event, entry, path),

            ('Battery', 'Specifications', 'Volt per cell crit'): lambda event, entry=entry, path=path:
                self.validate_cell_voltage(event, entry, path),

            ('Battery', 'Specifications', 'Number of cells'): lambda event, entry=entry, path=path:
                self.validate_entry_limits(event, entry, int, (1, 50), "Nr of cells", path),

            ('Battery', 'Specifications', 'Capacity mAh'): lambda event, entry=entry, path=path:
                self.validate_entry_limits(event, entry, int, (100, 1000000), "mAh capacity", path),

            ('Motors', 'Specifications', 'Poles'): lambda event, entry=entry, path=path:
                self.validate_entry_limits(event, entry, int, (3, 50), "Motor Poles", path),

            ('Propellers', 'Specifications', 'Diameter_inches'): lambda event, entry=entry, path=path:
                self.validate_entry_limits(event, entry, float, (0.3, 400), "Propeller Diameter", path),

        }
        return validate_functions.get(path, None)

    def validate_combobox(self, event, path) -> bool:
        """
        Validates the value of a combobox.
        """
        combobox = event.widget # Get the combobox widget that triggered the event
        value = combobox.get() # Get the current value of the combobox
        allowed_values = combobox.cget("values") # Get the list of allowed values

        if value not in allowed_values:
            if event.type == "10": # FocusOut events
                show_error_message("Error", f"Invalid value '{value}' for {'>'.join(list(path))}\n"
                                   f"Allowed values are: {', '.join(allowed_values)}")
            combobox.configure(style="comb_input_invalid.TCombobox")
            return False
        combobox.configure(style="comb_input_valid.TCombobox")
        return True

    def validate_entry_limits(self, event, entry, data_type, limits, name, path):  # pylint: disable=too-many-arguments
        is_focusout_event = event and event.type == "10"
        try:
            value = data_type(entry.get())
            if value < limits[0] or value > limits[1]:
                entry.configure(style="entry_input_invalid.TEntry")
                raise ValueError(f"{name} must be a {data_type.__name__} between {limits[0]} and {limits[1]}")
        except ValueError as e:
            if is_focusout_event:
                show_error_message("Error", f"Invalid value '{value}' for {'>'.join(list(path))}\n{e}")
            return False
        entry.configure(style="entry_input_valid.TEntry")
        return True

    def validate_cell_voltage(self, event, entry, path):  # pylint: disable=too-many-branches
        """
        Validates the value of a battery cell voltage entry.
        """
        chemistry_path = ('Battery', 'Specifications', 'Chemistry')
        if chemistry_path not in self.entry_widgets:
            show_error_message("Error", "Battery Chemistry not set. Will default to Lipo.")
            chemistry = "Lipo"
        else:
            chemistry = self.entry_widgets[chemistry_path].get()
        value = entry.get()
        is_focusout_event = event and event.type == "10"
        try:
            voltage = float(value)
            if voltage < BatteryCell.limit_min_voltage(chemistry):
                if is_focusout_event:
                    entry.delete(0, tk.END)
                    entry.insert(0, BatteryCell.limit_min_voltage(chemistry))
                raise VoltageTooLowError(f"is below the {chemistry} minimum limit of "
                                         f"{BatteryCell.limit_min_voltage(chemistry)}")
            if voltage > BatteryCell.limit_max_voltage(chemistry):
                if is_focusout_event:
                    entry.delete(0, tk.END)
                    entry.insert(0, BatteryCell.limit_max_voltage(chemistry))
                raise VoltageTooHighError(f"is above the {chemistry} maximum limit of "
                                          f"{BatteryCell.limit_max_voltage(chemistry)}")
        except (VoltageTooLowError, VoltageTooHighError) as e:
            if is_focusout_event:
                show_error_message("Error", f"Invalid value '{value}' for {'>'.join(list(path))}\n"
                                   f"{e}")
            else:
                entry.configure(style="entry_input_invalid.TEntry")
                return False
        except ValueError as e:
            if is_focusout_event:
                show_error_message("Error", f"Invalid value '{value}' for {'>'.join(list(path))}\n"
                                f"{e}\nWill be set to the recommended value.")
                entry.delete(0, tk.END)
                if path[-1] == "Volt per cell max":
                    entry.insert(0, str(BatteryCell.recommended_max_voltage(chemistry)))
                elif path[-1] == "Volt per cell low":
                    entry.insert(0, str(BatteryCell.recommended_low_voltage(chemistry)))
                elif path[-1] == "Volt per cell crit":
                    entry.insert(0, str(BatteryCell.recommended_crit_voltage(chemistry)))
                else:
                    entry.insert(0, "3.8")
            else:
                entry.configure(style="entry_input_invalid.TEntry")
                return False
        entry.configure(style="entry_input_valid.TEntry")
        return True

    def save_data(self):
        if self.validate_data():
            ComponentEditorWindowBase.save_data(self)

    def validate_data(self):  # pylint: disable=too-many-branches
        invalid_values = False
        duplicated_connections = False
        fc_serial_connection = {}

        for path, entry in self.entry_widgets.items():
            value = entry.get()

            if isinstance(entry, ttk.Combobox):
                if value not in entry.cget("values"):
                    show_error_message("Error", f"Invalid value '{value}' for {'>'.join(list(path))}\n"
                                    f"Allowed values are: {', '.join(entry.cget('values'))}")
                    entry.configure(style="comb_input_invalid.TCombobox")
                    invalid_values = True
                    continue
                if 'FC Connection' in path and 'Type' in path:
                    if value in fc_serial_connection and value not in ["CAN1", "CAN2", "I2C1", "I2C2", "I2C3", "I2C4"]:
                        if path[0] in ['Telemetry', 'RC Receiver'] and \
                           fc_serial_connection[value] in ['Telemetry', 'RC Receiver']:
                            entry.configure(style="comb_input_valid.TCombobox")
                            continue  # Allow telemetry and RC Receiver connections using the same SERIAL port
                        show_error_message("Error", f"Duplicate FC connection type '{value}' for {'>'.join(list(path))}")
                        entry.configure(style="comb_input_invalid.TCombobox")
                        duplicated_connections = True
                        continue
                    fc_serial_connection[value] = path[0]
                entry.configure(style="comb_input_valid.TCombobox")

            validate_function = self.get_validate_function(entry, path)
            if validate_function:
                mock_focus_out_event = type('', (), {'type': '10'})()
                if not validate_function(mock_focus_out_event):
                    invalid_values = True
            if path in [('Battery', 'Specifications', 'Volt per cell max'), ('Battery', 'Specifications', 'Volt per cell low'),
                        ('Battery', 'Specifications', 'Volt per cell crit')]:
                if not self.validate_cell_voltage(None, entry, path):
                    invalid_values = True
            if path == ('Battery', 'Specifications', 'Volt per cell low'):
                if value >= self.entry_widgets[('Battery', 'Specifications', 'Volt per cell max')].get():
                    show_error_message("Error", "Battery Cell Low voltage must be lower than max voltage")
                    entry.configure(style="entry_input_invalid.TEntry")
                    invalid_values = True
            if path == ('Battery', 'Specifications', 'Volt per cell crit'):
                if value >= self.entry_widgets[('Battery', 'Specifications', 'Volt per cell low')].get():
                    show_error_message("Error", "Battery Cell Crit voltage must be lower than low voltage")
                    entry.configure(style="entry_input_invalid.TEntry")
                    invalid_values = True

        return not (invalid_values or duplicated_connections)


if __name__ == "__main__":
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format='%(asctime)s - %(levelname)s - %(message)s')

    filesystem = LocalFilesystem(args.vehicle_dir, args.vehicle_type, None, args.allow_editing_template_files)
    app = ComponentEditorWindow(VERSION, filesystem)
    app.root.mainloop()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

class TemplateOverview:  # pylint: disable=too-many-instance-attributes
    """
    Represents a single vehicle template configuration within the ArduPilot Methodic Configurator.

    This class encapsulates the data and attributes associated with a specific vehicle template configuration.
    It is designed to hold information about various components of a drone, such as the flight controller, telemetry system,
    ESCs, propellers, and GNSS receiver, along with their specifications. The class facilitates easy access to these
    attributes, enabling the GUI to display and select the templates in a structured format.
    """
    def __init__(self, components_data: dict):
        # The declaration order of these parameters determines the column order in the GUI
        self.fc_manufacturer = components_data.get('Flight Controller', {}).get('Product', {}).get('Manufacturer', '')
        self.fc_model = components_data.get('Flight Controller', {}).get('Product', {}).get('Model', '')
        self.tow_max_kg = components_data.get('Frame', {}).get('Specifications', {}).get('TOW max Kg', '')
        self.prop_diameter_inches = components_data.get('Propellers', {}).get('Specifications', {}).get('Diameter_inches', '')
        self.rc_protocol = components_data.get('RC Receiver', {}).get('FC Connection', {}).get('Protocol', '')
        self.telemetry_model = components_data.get('Telemetry', {}).get('Product', {}).get('Model', '')
        self.esc_protocol = components_data.get('ESC', {}).get('FC Connection', {}).get('Protocol', '')
        self.gnss_model = components_data.get('GNSS receiver', {}).get('Product', {}).get('Model', '')
        self.gnss_connection = components_data.get('GNSS receiver', {}).get('FC Connection', {}).get('Type', '')

    @staticmethod
    def columns():
        # Must match the order in the __init__() function above
        return ("Template path",
                "FC\nManufacturer",
                "FC\nModel",
                "TOW Max\n[KG]",
                "Prop Diameter\n[inches]",
                "RC\nProtocol",
                "Telemetry\nModel",
                "ESC\nProtocol",
                "GNSS\nModel",
                "GNSS\nConnection",
                )

    def attributes(self):
        return self.__dict__.keys()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from os import path as os_path
from os import walk as os_walk

# from sys import exit as sys_exit
# from logging import debug as logging_debug
#from logging import info as logging_info
from logging import warning as logging_warning
from logging import error as logging_error

from re import match as re_match

from json import load as json_load
from json import dump as json_dump
from json import JSONDecodeError

from MethodicConfigurator.backend_filesystem_program_settings import ProgramSettings

from MethodicConfigurator.middleware_template_overview import TemplateOverview


class VehicleComponents:
    """
    This class provides methods to load and save
    vehicle components configurations from a JSON file.
    """
    def __init__(self):
        self.vehicle_components_json_filename = "vehicle_components.json"
        self.vehicle_components = None

    def load_vehicle_components_json_data(self, vehicle_dir: str):
        data = {}
        try:
            filepath = os_path.join(vehicle_dir, self.vehicle_components_json_filename)
            with open(filepath, 'r', encoding='utf-8') as file:
                data = json_load(file)
        except FileNotFoundError:
            logging_warning("File '%s' not found in %s.", self.vehicle_components_json_filename, vehicle_dir)
        except JSONDecodeError:
            logging_error("Error decoding JSON data from file '%s'.", filepath)
        self.vehicle_components = data
        return data

    def save_vehicle_components_json_data(self, data, vehicle_dir: str) -> bool:
        filepath = os_path.join(vehicle_dir, self.vehicle_components_json_filename)
        try:
            with open(filepath, 'w', encoding='utf-8') as file:
                json_dump(data, file, indent=4)
        except Exception as e:  # pylint: disable=broad-except
            logging_error("Error saving JSON data to file '%s': %s", filepath, e)
            return True
        return False

    def get_fc_fw_version_from_vehicle_components_json(self) -> str:
        if self.vehicle_components and 'Components' in self.vehicle_components:
            components = self.vehicle_components['Components']
        else:
            components = None
        if components:
            version_str = components.get('Flight Controller', {}).get('Firmware', {}).get('Version', '')
            version_str = version_str.lstrip().split(' ')[0] if version_str else ''
            if re_match(r'^\d+\.\d+\.\d+$', version_str):
                return version_str
            logging_error(f"FW version string {version_str} on {self.vehicle_components_json_filename} is invalid")
        return None

    @staticmethod
    def get_vehicle_components_overviews():
        """
        Finds all subdirectories of base_dir containing a "vehicle_components.json" file,
        creates a dictionary where the keys are the subdirectory names (relative to base_dir)
        and the values are instances of VehicleComponents.

        :param base_dir: The base directory to start searching from.
        :return: A dictionary mapping subdirectory paths to VehicleComponents instances.
        """
        vehicle_components_dict = {}
        file_to_find = VehicleComponents().vehicle_components_json_filename
        template_default_dir = ProgramSettings.get_templates_base_dir()
        for root, _dirs, files in os_walk(template_default_dir):
            if file_to_find in files:
                relative_path = os_path.relpath(root, template_default_dir)
                vehicle_components = VehicleComponents()
                comp_data = vehicle_components.load_vehicle_components_json_data(root)
                if comp_data:
                    comp_data = comp_data.get('Components', {})
                    vehicle_components_overview = TemplateOverview(comp_data)
                    vehicle_components_dict[relative_path] = vehicle_components_overview

        return vehicle_components_dict



#!/usr/bin/python3

'''
Extracts parameter values or parameter default values from an ArduPilot .bin log file.

Supports Mission Planner, MAVProxy and QGCS file format output

Currently has 95% unit test coverage

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

import argparse
import re
from typing import Dict, Tuple
from pymavlink import mavutil

NO_DEFAULT_VALUES_MESSAGE = "The .bin file contained no parameter default values. Update to a newer ArduPilot firmware version"
PARAM_NAME_REGEX = r'^[A-Z][A-Z_0-9]*$'
PARAM_NAME_MAX_LEN = 16
MAVLINK_SYSID_MAX = 2**24
MAVLINK_COMPID_MAX = 2**8
MAV_PARAM_TYPE_REAL32 = 9


def parse_arguments(args=None):
    """
    Parses command line arguments for the script.

    Args:
        args: List of command line arguments. Defaults to None, which means it uses sys.argv.

    Returns:
        Namespace object containing the parsed arguments.
    """
    parser = argparse.ArgumentParser(description='Extracts parameter default values from an ArduPilot .bin log file.')
    parser.add_argument('-f', '--format',
                        choices=['missionplanner', 'mavproxy', 'qgcs'],
                        default='missionplanner', help='Output file format. Defaults to %(default)s.',
                        )
    parser.add_argument('-s', '--sort',
                        choices=['none', 'missionplanner', 'mavproxy', 'qgcs'],
                        default='', help='Sort the parameters in the file. Defaults to the same as the format.',
                        )
    parser.add_argument('-v', '--version', action='version', version='%(prog)s 1.1',
                        help='Display version information and exit.',
                        )
    parser.add_argument('-i', '--sysid', type=int, default=-1,
                        help='System ID for qgcs output format. Defaults to SYSID_THISMAV if defined else 1.',
                        )
    parser.add_argument('-c', '--compid', type=int, default=-1,
                        help='Component ID for qgcs output format. Defaults to 1.',
                        )
    parser.add_argument('-t', '--type',
                        choices=['defaults', 'values', 'non_default_values'],
                        default='defaults', help='Type of parameter values to extract. Defaults to %(default)s.',
                        )
    parser.add_argument('bin_file', help='The ArduPilot .bin log file to read')
    args, _ = parser.parse_known_args(args)

    if args.sort == '':
        args.sort = args.format

    if args.format != 'qgcs':
        if args.sysid != -1:
            raise SystemExit("--sysid parameter is only relevant if --format is qgcs")
        if args.compid != -1:
            raise SystemExit("--compid parameter is only relevant if --format is qgcs")

    return args


def extract_parameter_values(logfile: str, param_type: str = 'defaults') -> Dict[str, float]:  # pylint: disable=too-many-branches
    """
    Extracts the parameter values from an ArduPilot .bin log file.

    Args:
        logfile: The path to the ArduPilot .bin log file.
        param_type: The type of parameter values to extract. Can be 'defaults', 'values' or 'non_default_values'.

    Returns:
        A dictionary with parameter names as keys and their values as float.
    """
    try:
        mlog = mavutil.mavlink_connection(logfile)
    except Exception as e:
        raise SystemExit(f"Error opening the {logfile} logfile: {str(e)}") from e
    values = {}
    while True:
        m = mlog.recv_match(type=['PARM'])
        if m is None:
            if not values:
                raise SystemExit(NO_DEFAULT_VALUES_MESSAGE)
            return values
        pname = m.Name
        if len(pname) > PARAM_NAME_MAX_LEN:
            raise SystemExit(f"Too long parameter name: {pname}")
        if not re.match(PARAM_NAME_REGEX, pname):
            raise SystemExit(f"Invalid parameter name {pname}")
        # parameter names are supposed to be unique
        if pname in values:
            continue
        if param_type == 'defaults':
            if hasattr(m, 'Default'):
                values[pname] = m.Default
        elif param_type == 'values':
            if hasattr(m, 'Value'):
                values[pname] = m.Value
        elif param_type == 'non_default_values':
            if hasattr(m, 'Value') and hasattr(m, 'Default') and m.Value != m.Default:
                values[pname] = m.Value
        else:
            raise SystemExit(f"Invalid type {param_type}")


def missionplanner_sort(item: str) -> Tuple[str, ...]:
    """
    Sorts a parameter name according to the rules defined in the Mission Planner software.

    Args:
        item: The parameter name to sort.

    Returns:
        A tuple representing the sorted parameter name.
    """
    parts = item.split("_")  # Split the parameter name by underscore
    # Compare the parts separately
    return tuple(parts)


def mavproxy_sort(item: str) -> str:
    """
    Sorts a parameter name according to the rules defined in the MAVProxy software.

    Args:
        item: The parameter name to sort.

    Returns:
        The sorted parameter name.
    """
    return item


def sort_params(params: Dict[str, float], sort_type: str = 'none') -> Dict[str, float]:
    """
    Sorts parameter names according to sort_type

    Args:
        params: A dictionary with parameter names as keys and their values as float.
        sort_type: The type of sorting to apply. Can be 'none', 'missionplanner', 'mavproxy' or 'qgcs'.

    Returns:
        A dictionary with parameter names as keys and their values as float.
    """
    if sort_type == "missionplanner":
        params = dict(sorted(params.items(), key=lambda x: missionplanner_sort(x[0])))
    elif sort_type == "mavproxy":
        params = dict(sorted(params.items(), key=lambda x: mavproxy_sort(x[0])))
    elif sort_type == "qgcs":
        params = {k: params[k] for k in sorted(params)}
    return params


def output_params(params: Dict[str, float], format_type: str = 'missionplanner',  # pylint: disable=too-many-branches
                  sysid: int = -1, compid: int = -1) -> None:
    """
    Outputs parameters names and their values to the console

    Args:
        params: A dictionary with parameter names as keys and their values as float.
        format_type: The output file format. Can be 'missionplanner', 'mavproxy' or 'qgcs'.

    Returns:
        None
    """
    if format_type == "qgcs":
        if sysid == -1:
            if 'SYSID_THISMAV' in params:
                sysid = int(params['SYSID_THISMAV'])
            else:
                sysid = 1  # if unspecified, default to 1
        if compid == -1:
            compid = 1  # if unspecified, default to 1
        if sysid < 0:
            raise SystemExit(f"Invalid system ID parameter {sysid} must not be negative")
        if sysid > MAVLINK_SYSID_MAX-1:
            raise SystemExit(f"Invalid system ID parameter {sysid} must be smaller than {MAVLINK_SYSID_MAX}")
        if compid < 0:
            raise SystemExit(f"Invalid component ID parameter {compid} must not be negative")
        if compid > MAVLINK_COMPID_MAX-1:
            raise SystemExit(f"Invalid component ID parameter {compid} must be smaller than {MAVLINK_COMPID_MAX}")
        # see https://dev.qgroundcontrol.com/master/en/file_formats/parameters.html
        print("""
# # Vehicle-Id Component-Id Name Value Type
""")

    for param_name, param_value in params.items():
        if format_type == "missionplanner":
            try:
                param_value = format(param_value, '.6f').rstrip('0').rstrip('.')
            except ValueError:
                pass # preserve non-floating point strings, if present
            print(f"{param_name},{param_value}")
        elif format_type == "mavproxy":
            print(f"{param_name:<15} {param_value:.6f}")
        elif format_type == "qgcs":
            print(f"{sysid} {compid} {param_name:<15} {param_value:.6f} {MAV_PARAM_TYPE_REAL32}")


def main():
    args = parse_arguments()
    parameter_values = extract_parameter_values(args.bin_file, args.type)
    parameter_values = sort_params(parameter_values, args.sort)
    output_params(parameter_values, args.format, args.sysid, args.compid)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

VERSION = '0.8.7'



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

import argparse
from logging import basicConfig as logging_basicConfig
from logging import getLevelName as logging_getLevelName
from logging import debug as logging_debug
from logging import info as logging_info
from logging import warning as logging_warning
from logging import error as logging_error
from sys import exit as sys_exit

from MethodicConfigurator.backend_filesystem import LocalFilesystem
from MethodicConfigurator.backend_flightcontroller import FlightController

from MethodicConfigurator.frontend_tkinter_base import show_error_message

from MethodicConfigurator.frontend_tkinter_connection_selection import ConnectionSelectionWindow

from MethodicConfigurator.frontend_tkinter_flightcontroller_info import FlightControllerInfoWindow

from MethodicConfigurator.frontend_tkinter_directory_selection import VehicleDirectorySelectionWindow

from MethodicConfigurator.frontend_tkinter_component_editor import ComponentEditorWindow

from MethodicConfigurator.frontend_tkinter_parameter_editor import ParameterEditorWindow

from MethodicConfigurator.common_arguments import add_common_arguments_and_parse

from MethodicConfigurator.version import VERSION


def argument_parser():
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.
    """
    parser = argparse.ArgumentParser(description='ArduPilot methodic configurator is a simple GUI with a table that lists '
                                     'parameters. The GUI reads intermediate parameter files from a directory and '
                                     'displays their parameters in a table. Each row displays the parameter name, '
                                     'its current value on the flight controller, its new value from the selected '
                                     'intermediate parameter file, and an "Upload" checkbox. The GUI includes "Upload '
                                     'selected params to FC" and "Skip" buttons at the bottom. '
                                     'When "Upload Selected to FC" is clicked, it uploads the selected parameters to the '
                                     'flight controller. '
                                     'When "Skip" is pressed, it skips to the next intermediate parameter file. '
                                     'The process gets repeated for each intermediate parameter file.')
    parser = FlightController.add_argparse_arguments(parser)
    parser = LocalFilesystem.add_argparse_arguments(parser)
    parser = ComponentEditorWindow.add_argparse_arguments(parser)
    return add_common_arguments_and_parse(parser)


def connect_to_fc_and_read_parameters(args):
    flight_controller = FlightController(args.reboot_time)

    error_str = flight_controller.connect(args.device)
    if error_str:
        if "No serial ports found" not in error_str:
            logging_error(error_str)
        conn_sel_window = ConnectionSelectionWindow(flight_controller, error_str)
        conn_sel_window.root.mainloop()

    vehicle_type = args.vehicle_type
    if vehicle_type == "":  # not explicitly set, to try to guess it
        if flight_controller.info.vehicle_type is not None:
            vehicle_type = flight_controller.info.vehicle_type
            logging_debug("Vehicle type not set explicitly, auto-detected %s.", vehicle_type)
    else:
        logging_info("Vehicle type explicitly set to %s.", vehicle_type)

    if vehicle_type == "": # did not guess it, default to ArduCopter
        vehicle_type = "ArduCopter"
        logging_warning("Could not detect vehicle type. Defaulting to ArduCopter.")
    return flight_controller,vehicle_type


def component_editor(args, flight_controller, vehicle_type, local_filesystem, vehicle_dir_window):
    component_editor_window = ComponentEditorWindow(VERSION, local_filesystem)
    component_editor_window.set_vehicle_type_and_version(vehicle_type, flight_controller.info.flight_sw_version_and_type)
    component_editor_window.set_fc_manufacturer(flight_controller.info.vendor)
    component_editor_window.set_fc_model(flight_controller.info.product)
    if vehicle_dir_window and \
       vehicle_dir_window.created_new_vehicle_from_template and \
       flight_controller.fc_parameters:
        # copy vehicle parameters to component editor values
        component_editor_window.set_values_from_fc_parameters(flight_controller.fc_parameters, local_filesystem.doc_dict)
    if args.skip_component_editor:
        component_editor_window.root.after(10, component_editor_window.root.destroy)
    component_editor_window.root.mainloop()

    if vehicle_dir_window and \
       vehicle_dir_window.created_new_vehicle_from_template and \
       vehicle_dir_window.use_fc_params.get():
        error_message = local_filesystem.copy_fc_params_values_to_template_created_vehicle_files(
            flight_controller.fc_parameters)
        if error_message:
            logging_error(error_message)
            show_error_message("Error in derived parameters", error_message)
            sys_exit(1)


def main():
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format='%(asctime)s - %(levelname)s - %(message)s')

    # Connect to the flight controller and read the parameters
    flight_controller, vehicle_type = connect_to_fc_and_read_parameters(args)

    if flight_controller.master is not None or args.device == 'test':
        FlightControllerInfoWindow(flight_controller)

    try:
        local_filesystem = LocalFilesystem(args.vehicle_dir, vehicle_type, flight_controller.info.flight_sw_version,
                                           args.allow_editing_template_files)
    except SystemExit as exp:
        show_error_message("Fatal error reading parameter files", f"{exp}")
        raise

    # Get the list of intermediate parameter files files that will be processed sequentially
    files = list(local_filesystem.file_parameters.keys())

    vehicle_dir_window = None
    if not files:
        vehicle_dir_window = VehicleDirectorySelectionWindow(local_filesystem, len(flight_controller.fc_parameters) > 0)
        vehicle_dir_window.root.mainloop()

    start_file = local_filesystem.get_start_file(args.n)

    component_editor(args, flight_controller, vehicle_type, local_filesystem, vehicle_dir_window)

    # Call the GUI function with the starting intermediate parameter file
    ParameterEditorWindow(start_file, flight_controller, local_filesystem, VERSION)

    # Close the connection to the flight controller
    flight_controller.disconnect()
    sys_exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
'''
decode ftp parameter protocol data
Original file was part of MAVProxy

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

import struct
import sys


class ParamData():
    """
    A class to manage parameter values and defaults for ArduPilot configuration.
    """
    def __init__(self):
        # params as (name, value, ptype)
        self.params = []
        # defaults as (name, value, ptype)
        self.defaults = None

    def add_param(self, name, value, ptype):
        self.params.append((name, value, ptype))

    def add_default(self, name, value, ptype):
        if self.defaults is None:
            self.defaults = []
        self.defaults.append((name, value, ptype))


def ftp_param_decode(data):  # pylint: disable=too-many-locals, too-many-branches
    '''decode parameter data, returning ParamData'''
    pdata = ParamData()

    magic = 0x671b
    magic_defaults = 0x671c
    if len(data) < 6:
        return None
    magic2, _num_params, total_params = struct.unpack("<HHH", data[0:6])
    if magic2 not in (magic, magic_defaults):
        print("paramftp: bad magic 0x%x expected 0x%x", magic2, magic)
        return None
    with_defaults = magic2 == magic_defaults
    data = data[6:]

    # mapping of data type to type length and format
    data_types = {
        1: (1, 'b'),
        2: (2, 'h'),
        3: (4, 'i'),
        4: (4, 'f'),
    }

    count = 0

    if sys.version_info.major < 3:
        pad_byte = chr(0)
        last_name = ''
    else:
        pad_byte = 0
        last_name = bytes()

    while True:
        while len(data) > 0 and data[0] == pad_byte:
            # skip pad bytes
            data = data[1:]

        if len(data) == 0:
            break

        ptype, plen = struct.unpack("<BB", data[0:2])
        flags = (ptype >> 4) & 0x0F
        has_default = with_defaults and (flags & 1) != 0
        ptype &= 0x0F

        if ptype not in data_types:
            print("paramftp: bad type 0x%x", ptype)
            return None

        (type_len, type_format) = data_types[ptype]
        default_len = type_len if has_default else 0

        name_len = ((plen >> 4) & 0x0F) + 1
        common_len = plen & 0x0F
        name = last_name[0:common_len] + data[2:2+name_len]
        vdata = data[2+name_len:2+name_len+type_len+default_len]
        last_name = name
        data = data[2+name_len+type_len+default_len:]
        if with_defaults:
            if has_default:
                v1, v2, = struct.unpack("<" + type_format + type_format, vdata)
                pdata.add_param(name, v1, ptype)
                pdata.add_default(name, v2, ptype)
            else:
                v, = struct.unpack("<" + type_format, vdata)
                pdata.add_param(name, v, ptype)
                pdata.add_default(name, v, ptype)
        else:
            v, = struct.unpack("<" + type_format, vdata)
            pdata.add_param(name, v, ptype)
        count += 1

    if count != total_params:
        print("paramftp: bad count %u should be %u", count, total_params)
        return None

    return pdata


def main():
    fname = sys.argv[1]
    with open(fname, 'rb') as file:
        data = file.read()
        print(f"Decoding file of length {len(data)}")
        pdata = ftp_param_decode(data)
        if pdata is None:
            print("Decode failed")
            sys.exit(1)
        for (name, value, _ptype) in pdata.params:
            print(name.decode('utf-8'), value)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from sys import exit as sys_exit

import tkinter as tk
from tkinter import messagebox
from tkinter import ttk
from tkinter import simpledialog

from logging import debug as logging_debug
from logging import info as logging_info
#from logging import warning as logging_warning
#from logging import error as logging_error
from logging import critical as logging_critical

from platform import system as platform_system

#from MethodicConfigurator.backend_filesystem import LocalFilesystem
from MethodicConfigurator.backend_filesystem import is_within_tolerance

#from MethodicConfigurator.backend_flightcontroller import FlightController

from MethodicConfigurator.frontend_tkinter_base import show_tooltip
#from MethodicConfigurator.frontend_tkinter_base import AutoResizeCombobox
from MethodicConfigurator.frontend_tkinter_base import ScrollFrame
from MethodicConfigurator.frontend_tkinter_base import get_widget_font

from MethodicConfigurator.frontend_tkinter_connection_selection import PairTupleCombobox

from MethodicConfigurator.annotate_params import Par


class ParameterEditorTable(ScrollFrame):  # pylint: disable=too-many-ancestors
    """
    A class to manage and display the parameter editor table within the GUI.

    This class inherits from ScrollFrame and is responsible for creating,
    managing, and updating the table that displays parameters for editing.
    """
    def __init__(self, root, local_filesystem, parameter_editor):
        super().__init__(root)
        self.root = root
        self.local_filesystem = local_filesystem
        self.parameter_editor = parameter_editor
        self.current_file = None
        self.upload_checkbutton_var = {}
        self.at_least_one_param_edited = False

        style = ttk.Style()
        style.configure('narrow.TButton', padding=0, width=4, border=(0, 0, 0, 0))

        # Prepare a dictionary that maps variable names to their values
        # These variables are used by the forced_parameters and derived_parameters in *_configuration_steps.json files
        self.variables = local_filesystem.get_eval_variables()

        self.compute_forced_and_derived_parameters()

    def compute_forced_and_derived_parameters(self):
        if self.local_filesystem.configuration_steps:
            for filename, file_info in self.local_filesystem.configuration_steps.items():
                error_msg = self.local_filesystem.compute_parameters(filename, file_info, 'forced', self.variables)
                if error_msg:
                    messagebox.showerror("Error in forced parameters", error_msg)
                #error_msg = self.local_filesystem.compute_parameters(filename, file_info, 'derived', self.variables)
                #if error_msg:
                #    messagebox.showerror("Error in derived parameters", error_msg)

    def repopulate(self, selected_file: str, different_params: dict, fc_parameters: dict, show_only_differences: bool):
        for widget in self.view_port.winfo_children():
            widget.destroy()
        self.current_file = selected_file

        # Create labels for table headers
        headers = ["-/+", "Parameter", "Current Value", "New Value", "Unit", "Upload", "Change Reason"]
        tooltips = ["Delete or add a parameter",
                    "Parameter name must be ^[A-Z][A-Z_0-9]* and most 16 characters long",
                    "Current value on the flight controller ",
                    "New value from the above selected intermediate parameter file",
                    "Parameter Unit",
                    "When selected, upload the new value to the flight controller",
                    "Reason why respective parameter changed"]
        for i, header in enumerate(headers):
            label = ttk.Label(self.view_port, text=header)
            label.grid(row=0, column=i, sticky="ew") # Use sticky="ew" to make the label stretch horizontally
            show_tooltip(label, tooltips[i])

        self.upload_checkbutton_var = {}

        # re-compute derived parameters because the fc_parameters values might have changed
        if self.local_filesystem.configuration_steps and selected_file in self.local_filesystem.configuration_steps:
            self.variables['fc_parameters'] = fc_parameters
            error_msg = self.local_filesystem.compute_parameters(selected_file,
                                                                 self.local_filesystem.configuration_steps[selected_file],
                                                                 'derived', self.variables)
            if error_msg:
                messagebox.showerror("Error in derived parameters", error_msg)

            self.rename_fc_connection(selected_file)

        if show_only_differences:
            self.__update_table(different_params, fc_parameters)
        else:
            self.__update_table(self.local_filesystem.file_parameters[selected_file], fc_parameters)
        # Scroll to the top of the parameter table
        self.canvas.yview("moveto", 0)

    def rename_fc_connection(self, selected_file):
        renames = {}
        if "rename_connection" in self.local_filesystem.configuration_steps[selected_file]:
            new_connection_prefix = self.local_filesystem.configuration_steps[selected_file]["rename_connection"]
            new_connection_prefix = eval(str(new_connection_prefix), {}, self.variables)  # pylint: disable=eval-used
            for param_name, _ in self.local_filesystem.file_parameters[selected_file].items():
                new_prefix = new_connection_prefix
                old_prefix = param_name.split("_")[0]

                # Handle CAN parameter names peculiarities
                if new_connection_prefix[:-1] == "CAN" and "CAN_P" in param_name:
                    old_prefix = param_name.split("_")[0] + '_' + param_name.split("_")[1]
                    new_prefix = "CAN_P" + new_connection_prefix[-1]
                if new_connection_prefix[:-1] == "CAN" and "CAN_D" in param_name:
                    old_prefix = param_name.split("_")[0] + '_' + param_name.split("_")[1]
                    new_prefix = "CAN_D" + new_connection_prefix[-1]

                if new_connection_prefix[:-1] in old_prefix:
                    renames[param_name] = param_name.replace(old_prefix, new_prefix)

        new_names = set()
        for old_name, new_name in renames.items():
            if new_name in new_names:
                self.local_filesystem.file_parameters[selected_file].pop(old_name)
                logging_info("Removing duplicate parameter %s", old_name)
            else:
                new_names.add(new_name)
                if new_name != old_name:
                    self.local_filesystem.file_parameters[selected_file][new_name] = \
                        self.local_filesystem.file_parameters[selected_file].pop(old_name)
                    logging_info("Renaming parameter %s to %s", old_name, new_name)

    def __update_table(self, params, fc_parameters):
        try:
            for i, (param_name, param) in enumerate(params.items(), 1):
                param_metadata = self.local_filesystem.doc_dict.get(param_name, None)
                param_default = self.local_filesystem.param_default_dict.get(param_name, None)
                doc_tooltip = param_metadata.get('doc_tooltip') if param_metadata else \
                    "No documentation available in apm.pdef.xml for this parameter"

                column = []
                column.append(self.__create_delete_button(param_name))
                column.append(self.__create_parameter_name(param_name, param_metadata, doc_tooltip))
                column.append(self.__create_flightcontroller_value(fc_parameters, param_name, param_default, doc_tooltip))
                column.append(self.__create_new_value_entry(param_name, param, param_metadata, param_default, doc_tooltip))
                column.append(self.__create_unit_label(param_metadata))
                column.append(self.__create_upload_checkbutton(param_name, bool(fc_parameters)))
                column.append(self.__create_change_reason_entry(param_name, param, column[3]))

                column[0].grid(row=i, column=0, sticky="w", padx=0)
                column[1].grid(row=i, column=1, sticky="w", padx=0)
                column[2].grid(row=i, column=2, sticky="e", padx=0)
                column[3].grid(row=i, column=3, sticky="e", padx=0)
                column[4].grid(row=i, column=4, sticky="e", padx=0)
                column[5].grid(row=i, column=5, sticky="e", padx=0)
                column[6].grid(row=i, column=6, sticky="ew", padx=(0, 5))

            # Add the "Add" button at the bottom of the table
            add_button = ttk.Button(self.view_port, text="Add", style='narrow.TButton',
                                    command=lambda: self.__on_parameter_add(fc_parameters))
            show_tooltip(add_button, f"Add a parameter to the {self.current_file} file")
            add_button.grid(row=len(params)+2, column=0, sticky="w", padx=0)


        except KeyError as e:
            logging_critical("Parameter %s not found in the %s file: %s", param_name, self.current_file, e, exc_info=True)
            sys_exit(1)

        # Configure the table_frame to stretch columns
        self.view_port.columnconfigure(0, weight=0) # Delete and Add buttons
        self.view_port.columnconfigure(1, weight=0, minsize=120) # Parameter name
        self.view_port.columnconfigure(2, weight=0) # Current Value
        self.view_port.columnconfigure(3, weight=0) # New Value
        self.view_port.columnconfigure(4, weight=0) # Units
        self.view_port.columnconfigure(5, weight=0) # Upload to FC
        self.view_port.columnconfigure(6, weight=1) # Change Reason

    def __create_delete_button(self, param_name):
        delete_button = ttk.Button(self.view_port, text="Del", style='narrow.TButton',
                                   command=lambda: self.__on_parameter_delete(param_name))
        show_tooltip(delete_button, f"Delete {param_name} from the {self.current_file} file")
        return delete_button

    def __create_parameter_name(self, param_name, param_metadata, doc_tooltip):
        is_calibration = param_metadata.get('Calibration', False) if param_metadata else False
        is_readonly = param_metadata.get('ReadOnly', False) if param_metadata else False
        parameter_label = ttk.Label(self.view_port, text=param_name + (" " * (16 - len(param_name))),
                                           background="red" if is_readonly else "yellow" if is_calibration else
                                           ttk.Style(self.root).lookup('TFrame', 'background'))
        if doc_tooltip:
            show_tooltip(parameter_label, doc_tooltip)
        return parameter_label

    def __create_flightcontroller_value(self, fc_parameters, param_name, param_default, doc_tooltip):
        if param_name in fc_parameters:
            value_str = format(fc_parameters[param_name], '.6f').rstrip('0').rstrip('.')
            if param_default is not None and is_within_tolerance(fc_parameters[param_name], param_default.value):
                # If it matches, set the background color to light blue
                flightcontroller_value = ttk.Label(self.view_port, text=value_str,
                                                          background="light blue")
            else:
                # Otherwise, set the background color to the default color
                flightcontroller_value = ttk.Label(self.view_port, text=value_str)
        else:
            flightcontroller_value = ttk.Label(self.view_port, text="N/A", background="orange")
        if doc_tooltip:
            show_tooltip(flightcontroller_value, doc_tooltip)
        return flightcontroller_value

    @staticmethod
    def __update_new_value_entry_text(new_value_entry: ttk.Entry, value: float, param_default):
        new_value_entry.delete(0, tk.END)
        value_str = format(value, '.6f').rstrip('0').rstrip('.')
        new_value_entry.insert(0, value_str)
        new_value_background = "light blue" if param_default is not None and \
            is_within_tolerance(value, param_default.value) else "white"
        new_value_entry.config(background=new_value_background)

    def __create_new_value_entry(self, param_name, param,  # pylint: disable=too-many-arguments
                                 param_metadata, param_default, doc_tooltip):

        present_as_forced = False
        if self.current_file in self.local_filesystem.forced_parameters and \
           param_name in self.local_filesystem.forced_parameters[self.current_file]:
            present_as_forced = True
            if not is_within_tolerance(param.value,
                                       self.local_filesystem.forced_parameters[self.current_file][param_name].value):
                param.value = self.local_filesystem.forced_parameters[self.current_file][param_name].value
                self.at_least_one_param_edited = True
        if self.current_file in self.local_filesystem.derived_parameters and \
           param_name in self.local_filesystem.derived_parameters[self.current_file]:
            present_as_forced = True
            if not is_within_tolerance(param.value,
                                       self.local_filesystem.derived_parameters[self.current_file][param_name].value):
                param.value = self.local_filesystem.derived_parameters[self.current_file][param_name].value
                self.at_least_one_param_edited = True

        bitmask_dict = None
        value_str = format(param.value, '.6f').rstrip('0').rstrip('.')
        if 'values' in param_metadata and param_metadata['values'] and \
           value_str in param_metadata['values']:
            selected_value = param_metadata['values'].get(value_str, None)
            new_value_entry = PairTupleCombobox(self.view_port, param_metadata['values'],
                                                value_str, param_name,
                                                style='TCombobox' if present_as_forced else 'readonly.TCombobox')
            new_value_entry.set(selected_value)
            font = get_widget_font(new_value_entry)
            font['size'] -= 2 if platform_system() == 'Windows' else 1
            new_value_entry.config(state='readonly', width=9, font=(font['family'], font['size']))
        else:
            new_value_entry = ttk.Entry(self.view_port, width=10, justify=tk.RIGHT)
            ParameterEditorTable.__update_new_value_entry_text(new_value_entry, param.value, param_default)
            bitmask_dict = param_metadata.get('Bitmask', None) if param_metadata else None
        try:
            old_value = self.local_filesystem.file_parameters[self.current_file][param_name].value
        except KeyError as e:
            logging_critical("Parameter %s not found in the %s file: %s", param_name, self.current_file, e, exc_info=True)
            sys_exit(1)
        if present_as_forced:
            new_value_entry.config(state='disabled', background='light grey')
        else:
            if bitmask_dict:
                new_value_entry.bind("<FocusIn>", lambda event:
                                    self.__open_bitmask_selection_window(event, param_name, bitmask_dict, old_value))
            else:
                new_value_entry.bind("<FocusOut>", lambda event, current_file=self.current_file, param_name=param_name:
                                        self.__on_parameter_value_change(event, current_file, param_name))
        if doc_tooltip:
            show_tooltip(new_value_entry, doc_tooltip)
        return new_value_entry

    def __open_bitmask_selection_window(self, event, param_name, bitmask_dict, old_value):  # pylint: disable=too-many-locals
        def on_close():
            checked_keys = [key for key, var in checkbox_vars.items() if var.get()]
            # Convert checked keys back to a decimal value
            new_decimal_value = sum(1 << key for key in checked_keys)
            # Update new_value_entry with the new decimal value
            ParameterEditorTable.__update_new_value_entry_text(event.widget,
                                                               new_decimal_value,
                                                               self.local_filesystem.param_default_dict.get(param_name, None))
            self.at_least_one_param_edited = (old_value != new_decimal_value) or self.at_least_one_param_edited
            self.local_filesystem.file_parameters[self.current_file][param_name].value = new_decimal_value
            # Destroy the window
            window.destroy()
            # Issue a FocusIn event on something else than new_value_entry to prevent endless looping
            self.root.focus_set()
            # Run the Tk event loop once to process the event
            self.root.update_idletasks()
            # Re-bind the FocusIn event to new_value_entry
            event.widget.bind("<FocusIn>", lambda event:
                                        self.__open_bitmask_selection_window(event, param_name, bitmask_dict, old_value))

        def update_label():
            checked_keys = [key for key, var in checkbox_vars.items() if var.get()]
            new_decimal_value = sum(1 << key for key in checked_keys)
            close_label.config(text=f"{param_name} Value: {new_decimal_value}")

        # Temporarily unbind the FocusIn event to prevent triggering the window again
        event.widget.unbind("<FocusIn>")
        window = tk.Toplevel(self.root)
        window.title(f"Select {param_name} Bitmask Options")
        checkbox_vars = {}

        main_frame = ttk.Frame(window)
        main_frame.pack(expand=True, fill=tk.BOTH)

        # Convert current_value to a set of checked keys
        current_value = int(event.widget.get())
        checked_keys = {key for key, _value in bitmask_dict.items() if (current_value >> key) & 1}

        for i, (key, value) in enumerate(bitmask_dict.items()):
            var = tk.BooleanVar(value=key in checked_keys)
            checkbox_vars[key] = var
            checkbox = ttk.Checkbutton(main_frame, text=value, variable=var, command=update_label)
            checkbox.grid(row=i, column=0, sticky="w")

        # Calculate new_decimal_value here to ensure it's accessible when creating the label
        new_decimal_value = sum(1 << key for key in checked_keys)

        # Replace the close button with a read-only label displaying the current new_decimal_value
        close_label = ttk.Label(main_frame, text=f"{param_name} Value: {new_decimal_value}")
        close_label.grid(row=len(bitmask_dict), column=0, pady=10)

        # Bind the on_close function to the window's WM_DELETE_WINDOW protocol
        window.protocol("WM_DELETE_WINDOW", on_close)

        # Make sure the window is visible before disabling the parent window
        window.deiconify()
        self.root.update_idletasks()
        window.grab_set()

        window.wait_window() # Wait for the window to be closed

    def __create_unit_label(self, param_metadata):
        unit_label = ttk.Label(self.view_port, text=param_metadata.get('unit') if param_metadata else "")
        unit_tooltip = param_metadata.get('unit_tooltip') if param_metadata else \
            "No documentation available in apm.pdef.xml for this parameter"
        if unit_tooltip:
            show_tooltip(unit_label, unit_tooltip)
        return unit_label

    def __create_upload_checkbutton(self, param_name, fc_connected):
        self.upload_checkbutton_var[param_name] = tk.BooleanVar(value=fc_connected)
        upload_checkbutton = ttk.Checkbutton(self.view_port, variable=self.upload_checkbutton_var[param_name])
        upload_checkbutton.configure(state='normal' if fc_connected else 'disabled')
        show_tooltip(upload_checkbutton, f'When selected upload {param_name} new value to the flight controller')
        return upload_checkbutton

    def __create_change_reason_entry(self, param_name, param, new_value_entry):

        present_as_forced = False
        if self.current_file in self.local_filesystem.forced_parameters and \
           param_name in self.local_filesystem.forced_parameters[self.current_file]:
            present_as_forced = True
            if param.comment != self.local_filesystem.forced_parameters[self.current_file][param_name].comment:
                param.comment = self.local_filesystem.forced_parameters[self.current_file][param_name].comment
                self.at_least_one_param_edited = True
        if self.current_file in self.local_filesystem.derived_parameters and \
           param_name in self.local_filesystem.derived_parameters[self.current_file]:
            present_as_forced = True
            if param.comment != self.local_filesystem.derived_parameters[self.current_file][param_name].comment:
                param.comment = self.local_filesystem.derived_parameters[self.current_file][param_name].comment
                self.at_least_one_param_edited = True

        change_reason_entry = ttk.Entry(self.view_port, background="white")
        change_reason_entry.insert(0, "" if param.comment is None else param.comment)
        if present_as_forced:
            change_reason_entry.config(state='disabled', background='light grey')
        else:
            change_reason_entry.bind("<FocusOut>", lambda event, current_file=self.current_file, param_name=param_name:
                                     self.__on_parameter_change_reason_change(event, current_file, param_name))
        show_tooltip(change_reason_entry, f'Reason why {param_name} should change to {new_value_entry.get()}')
        return change_reason_entry

    def __on_parameter_delete(self, param_name):
        if messagebox.askyesno(f"{self.current_file}", f"Are you sure you want to delete the {param_name} parameter?"):
            del self.local_filesystem.file_parameters[self.current_file][param_name]
            self.at_least_one_param_edited = True
            self.parameter_editor.repopulate_parameter_table(self.current_file)

    def __on_parameter_add(self, fc_parameters):
        # Prompt the user for a parameter name
        param_name = simpledialog.askstring("New parameter name", "Enter new parameter name:")
        if not param_name:
            messagebox.showerror("Parameter name can not be empty.")
            return
        if param_name in self.local_filesystem.file_parameters[self.current_file]:
            messagebox.showerror("Parameter already exists, edit it instead")
            return
        if fc_parameters:
            if param_name in fc_parameters:
                self.local_filesystem.file_parameters[self.current_file][param_name] = Par(fc_parameters[param_name], "")
                self.at_least_one_param_edited = True
                self.parameter_editor.repopulate_parameter_table(self.current_file)
            else:
                messagebox.showerror("Invalid parameter name.", "Parameter name not found in the flight controller.")
        elif self.local_filesystem.doc_dict:
            if param_name in self.local_filesystem.doc_dict:
                self.local_filesystem.file_parameters[self.current_file][param_name] = Par( \
                    self.local_filesystem.param_default_dict.get(param_name, Par(0, "")).value, "")
                self.at_least_one_param_edited = True
                self.parameter_editor.repopulate_parameter_table(self.current_file)
            else:
                messagebox.showerror("Invalid parameter name.", "Parameter name not found in the apm.pdef.xml file.")
        else:
            messagebox.showerror("Operation not possible",
                                    "Can not add parameter when no FC is connected and no apm.pdef.xml file exists.")


    def __on_parameter_value_change(self, event, current_file, param_name):
        # Get the new value from the Entry widget
        if isinstance(event.widget, PairTupleCombobox):
            new_value = event.widget.get_selected_key()
        else:
            new_value = event.widget.get()
        try:
            old_value = self.local_filesystem.file_parameters[current_file][param_name].value
        except KeyError as e:
            logging_critical("Parameter %s not found in the %s file: %s", param_name, current_file, e, exc_info=True)
            sys_exit(1)
        valid = True
        # Check if the input is a valid float
        try:
            p = float(new_value)
            changed = not is_within_tolerance(old_value, p)
            param_metadata = self.local_filesystem.doc_dict.get(param_name, None)
            p_min = param_metadata.get('min', None) if param_metadata else None
            p_max = param_metadata.get('max', None) if param_metadata else None
            if changed:
                if p_min and p < p_min:
                    if not messagebox.askyesno("Out-of-bounds Value",
                                               f"The value for {param_name} {p} should be greater than {p_min}\n"
                                               "Use out-of-bounds value?", icon='warning'):
                        valid = False
                if p_max and p > p_max:
                    if not messagebox.askyesno("Out-of-bounds Value",
                                               f"The value for {param_name} {p} should be smaller than {p_max}\n"
                                               "Use out-of-bounds value?", icon='warning'):
                        valid = False
        except ValueError:
            # Optionally, you can handle the invalid value here, for example, by showing an error message
            messagebox.showerror("Invalid Value", f"The value for {param_name} must be a valid float.")
            valid = False
        if valid:
            if changed and not self.at_least_one_param_edited:
                logging_debug("Parameter %s changed, will later ask if change(s) should be saved to file.", param_name)
            self.at_least_one_param_edited = changed or self.at_least_one_param_edited
            # Update the params dictionary with the new value
            self.local_filesystem.file_parameters[current_file][param_name].value = p
        else:
            # Revert to the previous (valid) value
            p = old_value
        ParameterEditorTable.__update_new_value_entry_text(event.widget, p,
                                                           self.local_filesystem.param_default_dict.get(param_name, None))

    def __on_parameter_change_reason_change(self, event, current_file, param_name):
        # Get the new value from the Entry widget
        new_value = event.widget.get()
        try:
            changed = new_value != self.local_filesystem.file_parameters[current_file][param_name].comment and \
                not (new_value == "" and self.local_filesystem.file_parameters[current_file][param_name].comment is None)
        except KeyError as e:
            logging_critical("Parameter %s not found in the %s file %s: %s", param_name, current_file,
                             new_value, e, exc_info=True)
            sys_exit(1)
        if changed and not self.at_least_one_param_edited:
            logging_debug("Parameter %s change reason changed from %s to %s, will later ask if change(s) should be saved to "
                          "file.",
                          param_name, self.local_filesystem.file_parameters[current_file][param_name].comment, new_value)
        self.at_least_one_param_edited = changed or self.at_least_one_param_edited
        # Update the params dictionary with the new value
        self.local_filesystem.file_parameters[current_file][param_name].comment = new_value

    def get_upload_selected_params(self, current_file: str):
        selected_params = {}
        for param_name, checkbutton_state in self.upload_checkbutton_var.items():
            if checkbutton_state.get():
                selected_params[param_name] = self.local_filesystem.file_parameters[current_file][param_name]
        return selected_params

    def generate_edit_widgets_focus_out(self):
        # Trigger the <FocusOut> event for all entry widgets to ensure all changes are processed
        for widget in self.view_port.winfo_children():
            if isinstance(widget, ttk.Entry):
                widget.event_generate("<FocusOut>", when="now")

    def get_at_least_one_param_edited(self):
        return self.at_least_one_param_edited

    def set_at_least_one_param_edited(self, value):
        self.at_least_one_param_edited = value



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from MethodicConfigurator.version import VERSION


def add_common_arguments_and_parse(parser):
    parser.add_argument('--loglevel',
                        type=str,
                        default='INFO',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help='Logging level (default is INFO).')
    parser.add_argument('-v', '--version',
                        action='version',
                        version=f'%(prog)s {VERSION}',
                        help='Display version information and exit.')
    return parser.parse_args()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from os import path as os_path
from os import getcwd as os_getcwd
from os import listdir as os_listdir
from os import rename as os_rename

from shutil import copy2 as shutil_copy2
from shutil import copytree as shutil_copytree

from re import compile as re_compile

# from sys import exit as sys_exit
from logging import debug as logging_debug
from logging import info as logging_info
from logging import warning as logging_warning
from logging import error as logging_error

from typing import Dict
from typing import List
from typing import Tuple

from zipfile import ZipFile

from MethodicConfigurator.annotate_params import PARAM_DEFINITION_XML_FILE, Par
from MethodicConfigurator.annotate_params import get_xml_url
from MethodicConfigurator.annotate_params import get_xml_dir
from MethodicConfigurator.annotate_params import parse_parameter_metadata
from MethodicConfigurator.annotate_params import format_columns
from MethodicConfigurator.annotate_params import split_into_lines
from MethodicConfigurator.annotate_params import update_parameter_documentation

from MethodicConfigurator.backend_filesystem_vehicle_components import VehicleComponents
from MethodicConfigurator.backend_filesystem_configuration_steps import ConfigurationSteps
from MethodicConfigurator.backend_filesystem_program_settings import ProgramSettings

TOOLTIP_MAX_LENGTH = 105


def is_within_tolerance(x: float, y: float, atol: float = 1e-08, rtol: float = 1e-03) -> bool:
    """
    Determines if the absolute difference between two numbers is within a specified tolerance.

    This function checks if the absolute difference between `x` and `y` is less than or equal to
    the sum of the absolute tolerance (`atol`) and the product of the relative tolerance (`rtol`)
    and the absolute value of `y`.

    Parameters:
    - x (float): The first number to compare.
    - y (float): The second number to compare.
    - atol (float, optional): The absolute tolerance. Defaults to 1e-08.
    - rtol (float, optional): The relative tolerance. Defaults to 1e-03.

    Returns:
    - bool: True if the difference is within the tolerance, False otherwise.
    """
    return abs(x - y) <= atol + (rtol * abs(y))


class LocalFilesystem(VehicleComponents, ConfigurationSteps, ProgramSettings):  # pylint: disable=too-many-public-methods
    """
    A class to manage local filesystem operations for the ArduPilot methodic configurator.

    This class provides methods for initializing and re-initializing the filesystem context,
    reading parameters from files, and handling configuration steps. It is designed to simplify
    the interaction with the local filesystem for managing ArduPilot configuration files.

    Attributes:
        vehicle_dir (str): The directory path where the vehicle configuration files are stored.
        vehicle_type (str): The type of the vehicle (e.g., "ArduCopter", "Rover").
        file_parameters (dict): A dictionary of parameters read from intermediate parameter files.
        param_default_dict (dict): A dictionary of default parameter values.
        doc_dict (dict): A dictionary containing documentation for each parameter.
    """
    def __init__(self, vehicle_dir: str, vehicle_type: str, fw_version: str, allow_editing_template_files: bool):
        self.file_parameters = None
        VehicleComponents.__init__(self)
        ConfigurationSteps.__init__(self, vehicle_dir, vehicle_type)
        ProgramSettings.__init__(self)
        self.fw_version = fw_version
        self.allow_editing_template_files = allow_editing_template_files
        if vehicle_dir is not None:
            self.re_init(vehicle_dir, vehicle_type)

    def re_init(self, vehicle_dir: str, vehicle_type: str):
        ConfigurationSteps.re_init(self, vehicle_dir, vehicle_type)
        self.vehicle_dir = vehicle_dir
        self.vehicle_type = vehicle_type
        self.param_default_dict = {}
        self.doc_dict = {}

        # Rename parameter files if some new files got added to the vehicle directory
        self.rename_parameter_files()

        # Read intermediate parameters from files
        self.file_parameters = self.read_params_from_files()
        if not self.file_parameters:
            return # No files intermediate parameters files found, no need to continue, the rest needs them

        self.load_vehicle_components_json_data(vehicle_dir)

        if self.fw_version is None:
            self.fw_version = self.get_fc_fw_version_from_vehicle_components_json()

        # Read ArduPilot parameter documentation
        xml_url = get_xml_url(vehicle_type, self.fw_version)
        xml_dir = get_xml_dir(vehicle_dir)
        [self.doc_dict, self.param_default_dict] = parse_parameter_metadata(xml_url, xml_dir, PARAM_DEFINITION_XML_FILE,
                                                                            vehicle_type, TOOLTIP_MAX_LENGTH)

        # Extend parameter documentation metadata if <parameter_file>.pdef.xml exists
        for filename in self.file_parameters:
            pdef_xml_file = filename.replace(".param", ".pdef.xml")
            if os_path.exists(os_path.join(xml_dir, pdef_xml_file)):
                [doc_dict, _param_default_dict] = parse_parameter_metadata("", xml_dir, pdef_xml_file,
                                                                           vehicle_type, TOOLTIP_MAX_LENGTH)
                self.doc_dict.update(doc_dict)

        self.__extend_and_reformat_parameter_documentation_metadata()

    def rename_parameter_files(self):
        # Rename parameter files if some new files got added to the vehicle directory
        if self.vehicle_dir is not None and self.configuration_steps is not None:
            for new_filename in self.configuration_steps:
                if 'old_filenames' in self.configuration_steps[new_filename]:
                    for old_filename in self.configuration_steps[new_filename]['old_filenames']:
                        if self.intermediate_parameter_file_exists(old_filename) and old_filename != new_filename:
                            new_filename_path = os_path.join(self.vehicle_dir, new_filename)
                            old_filename_path = os_path.join(self.vehicle_dir, old_filename)
                            os_rename(old_filename_path, new_filename_path)
                            logging_info("Renamed %s to %s", old_filename, new_filename)

    def __extend_and_reformat_parameter_documentation_metadata(self):  # pylint: disable=too-many-branches
        for param_name, param_info in self.doc_dict.items():
            if 'fields' in param_info:
                param_fields = param_info['fields']
                if 'Units' in param_fields:
                    units_list = param_fields['Units'].split('(')
                    param_info['unit'] = units_list[0].strip()
                    if len(units_list) > 1:
                        param_info['unit_tooltip'] = units_list[1].strip(')').strip()
                if 'Range' in param_fields:
                    param_info['min'] = float(param_fields['Range'].split(' ')[0].strip())
                    param_info['max'] = float(param_fields['Range'].split(' ')[1].strip())
                if 'Calibration' in param_fields:
                    param_info['Calibration'] = self.str_to_bool(param_fields['Calibration'].strip())
                if 'ReadOnly' in param_fields:
                    param_info['ReadOnly'] = self.str_to_bool(param_fields['ReadOnly'].strip())
                if 'RebootRequired' in param_fields:
                    param_info['RebootRequired'] = self.str_to_bool(param_fields['RebootRequired'].strip())
                if 'Bitmask' in param_fields:
                    bitmask_items = param_fields['Bitmask'].split(',')
                    param_info['Bitmask'] = {}
                    for item in bitmask_items:
                        key, value = item.split(':')
                        param_info['Bitmask'][int(key.strip())] = value.strip()

            if 'values' in param_info and param_info['values']:
                try:
                    param_info['Values'] = {int(k): v for k, v in param_info['values'].items()}
                except ValueError:
                    param_info['Values'] = {float(k): v for k, v in param_info['values'].items()}
                # print(param_info['Values'])

            prefix_parts = [
                f"{param_info['humanName']}",
            ]
            prefix_parts += param_info["documentation"]
            for key, value in param_info["fields"].items():
                if key not in ['Units', 'UnitText']:
                    prefix_parts += split_into_lines(f"{key}: {value}", TOOLTIP_MAX_LENGTH)
            prefix_parts += format_columns(param_info["values"], TOOLTIP_MAX_LENGTH)
            if param_name in self.param_default_dict:
                default_value = format(self.param_default_dict[param_name].value, '.6f').rstrip('0').rstrip('.')
                prefix_parts += [f"Default: {default_value}"]
            param_info['doc_tooltip'] = ('\n').join(prefix_parts)

    def read_params_from_files(self):
        """
        Reads intermediate parameter files from a directory and stores their contents in a dictionary.

        This function scans the specified directory for files matching a specific pattern,
        reads each file, and stores the parameter names and values in a dictionary.
        Files named '00_default.param' and '01_ignore_readonly.param' are ignored.

        Returns:
        - Dict[str, Dict[str, 'Par']]: A dictionary with filenames as keys and as values
                                       a dictionary with (parameter names, values) pairs.
        """
        parameters = {}
        if os_path.isdir(self.vehicle_dir):
            # Regular expression pattern for filenames starting with two digits followed by an underscore and ending in .param
            pattern = re_compile(r'^\d{2}_.*\.param$')

            for filename in sorted(os_listdir(self.vehicle_dir)):
                if pattern.match(filename):
                    if filename in ['00_default.param', '01_ignore_readonly.param']:
                        continue
                    parameters[filename] = Par.load_param_file_into_dict(os_path.join(self.vehicle_dir, filename))
        else:
            logging_error("Error: %s is not a directory.", self.vehicle_dir)
        return parameters

    @staticmethod
    def str_to_bool(s):
        """
        Converts a string representation of a boolean value to a boolean.

        This function interprets the string 'true', 'yes', '1' as True, and 'false', 'no', '0' as False.
        Any other input will return None.

        Parameters:
        - s (str): The string to convert.

        Returns:
        - Optional[bool]: True, False, or None if the string does not match any known boolean representation.
        """
        if s.lower() == "true" or s.lower() == "yes" or s.lower() == "1":
            return True
        if s.lower() == "false" or s.lower() == "no" or s.lower() == "0":
            return False
        return None

    def export_to_param(self, params: Dict[str, 'Par'], filename_out: str, annotate_doc: bool = True) -> None:
        """
        Exports a dictionary of parameters to a .param file and optionally annotates the documentation.

        This function formats the provided parameters into a string suitable for a .param file,
        writes the string to the specified output file, and optionally updates the parameter documentation.

        Parameters:
        - params (Dict[str, 'Par']): A dictionary of parameters to export.
        - filename_out (str): The name of the output file.
        - annotate_doc (bool, optional): Whether to update the parameter documentation. Defaults to True.
        """
        Par.export_to_param(Par.format_params(params), os_path.join(self.vehicle_dir, filename_out))
        if annotate_doc:
            update_parameter_documentation(self.doc_dict,
                                           os_path.join(self.vehicle_dir, filename_out),
                                           "missionplanner",
                                           self.param_default_dict)

    def intermediate_parameter_file_exists(self, filename: str) -> bool:
        """
        Check if an intermediate parameter file exists in the vehicle directory.

        Parameters:
        - filename (str): The name of the file to check.

        Returns:
        - bool: True if the file exists and is a file (not a directory), False otherwise.
        """
        return os_path.exists(os_path.join(self.vehicle_dir, filename)) and \
            os_path.isfile(os_path.join(self.vehicle_dir, filename))

    def __all_intermediate_parameter_file_comments(self) -> Dict[str, str]:
        """
        Retrieves all comments associated with parameters from intermediate parameter files.

        This method iterates through all intermediate parameter files, collects comments for each parameter,
        and returns them as a dictionary where the keys are parameter names and the values are the comments.
        Comments from the same parameter in different files are not merged; only the comment from the last file is returned.

        Returns:
        - Dict[str, str]: A dictionary mapping parameter names to their comments.
        """
        ret = {}
        for _filename, params in self.file_parameters.items():
            for param, info in params.items():
                if info.comment:
                    ret[param] = info.comment
        return ret

    def annotate_intermediate_comments_to_param_dict(self, param_dict: Dict[str, float]) -> Dict[str, 'Par']:
        """
        Annotates comments from intermediate parameter files to a parameter value-only dictionary.

        This function takes a dictionary of parameters with only values and adds comments from
        intermediate parameter files to create a new dictionary where each parameter is represented
        by a 'Par' object containing both the value and the comment.

        Parameters:
        - param_dict (Dict[str, float]): A dictionary of parameters with only values.

        Returns:
        - Dict[str, 'Par']: A dictionary of parameters with intermediate parameter file comments.
        """
        ret = {}
        ip_comments = self.__all_intermediate_parameter_file_comments()
        for param, value in param_dict.items():
            ret[param] = Par(float(value), ip_comments.get(param, ''))
        return ret

    def categorize_parameters(self, param: Dict[str, 'Par']) -> List[Dict[str, 'Par']]:
        """
        Categorize parameters into three categories based on their default values and documentation attributes.

        This method iterates through the provided dictionary of parameters and categorizes them into three groups:
        - Non-default, read-only parameters
        - Non-default, writable calibrations
        - Non-default, writable non-calibrations

        Parameters:
        - param (Dict[str, 'Par']): A dictionary mapping parameter names to their 'Par' objects.

        Returns:
        - List[Dict[str, 'Par']]: A list containing three dictionaries.
                                  Each dictionary represents one of the categories mentioned above.
        """
        non_default__read_only_params = {}
        non_default__writable_calibrations = {}
        non_default__writable_non_calibrations = {}
        for param_name, param_info in param.items():
            if param_name in self.param_default_dict and is_within_tolerance(param_info.value,
                                                                             self.param_default_dict[param_name].value):
                continue     # parameter has a default value, ignore it

            if param_name in self.doc_dict and self.doc_dict[param_name].get('ReadOnly', False):
                non_default__read_only_params[param_name] = param_info
                continue

            if param_name in self.doc_dict and self.doc_dict[param_name].get('Calibration', False):
                non_default__writable_calibrations[param_name] = param_info
                continue
            non_default__writable_non_calibrations[param_name] = param_info

        return non_default__read_only_params, non_default__writable_calibrations, non_default__writable_non_calibrations

    # Extract the vehicle name from the directory path
    def get_vehicle_directory_name(self):
        # Normalize the path to ensure it's in a standard format
        normalized_path = os_path.normpath(self.vehicle_dir)

        # Split the path into head and tail, then get the basename of the tail
        directory_name = os_path.basename(os_path.split(normalized_path)[1])

        return directory_name

    def zip_file_path(self):
        vehicle_name = self.get_vehicle_directory_name()
        return os_path.join(self.vehicle_dir, f"{vehicle_name}.zip")

    def zip_file_exists(self):
        zip_file_path = self.zip_file_path()
        return os_path.exists(zip_file_path) and os_path.isfile(zip_file_path)

    def zip_files(self, files_to_zip: List[Tuple[bool, str]]):
        """
        Zips the intermediate parameter files that were written to, including specific summary files.

        This method creates a zip archive containing all intermediate parameter files, along with
        specific summary files if they were written. The zip file is saved in the same directory as the
        intermediate parameter files. The method checks for the existence of each file before
        attempting to add it to the zip archive.

        Parameters:
        - files_to_zip (List[Tuple[bool, str]]): A list of tuples, where each tuple contains a boolean
                                            indicating if the file was written and a string for the filename.
        """
        zip_file_path = self.zip_file_path()
        with ZipFile(zip_file_path, 'w') as zipf:
            # Add all intermediate parameter files
            for file_name in self.file_parameters:
                zipf.write(os_path.join(self.vehicle_dir, file_name), arcname=file_name)

            # Check for and add specific files if they exist
            specific_files = ["00_default.param", "apm.pdef.xml", self.configuration_steps_filename,
                              "vehicle_components.json", "vehicle.jpg", "last_uploaded_filename.txt",
                              "tempcal_gyro.png", "tempcal_acc.png"]
            for file_name in specific_files:
                file_path = os_path.join(self.vehicle_dir, file_name)
                if os_path.exists(file_path):
                    zipf.write(file_path, arcname=file_name)

            for wrote, filename in files_to_zip:
                if wrote:
                    file_path = os_path.join(self.vehicle_dir, filename)
                    if os_path.exists(file_path):
                        zipf.write(file_path, arcname=filename)

        logging_info("Intermediate parameter files and summary files zipped to %s", zip_file_path)

    def vehicle_image_filepath(self):
        return os_path.join(self.vehicle_dir, 'vehicle.jpg')

    def vehicle_image_exists(self):
        return os_path.exists(self.vehicle_image_filepath())

    @staticmethod
    def new_vehicle_dir(base_dir: str, new_dir: str):
        return os_path.join(base_dir, new_dir)

    @staticmethod
    def directory_exists(directory: str) -> bool:
        return os_path.exists(directory)

    def copy_template_files_to_new_vehicle_dir(self, template_dir: str, new_vehicle_dir: str):
        # Copy the template files to the new vehicle directory
        for item in os_listdir(template_dir):
            if item == 'vehicle.jpg':
                continue
            s = os_path.join(template_dir, item)
            d = os_path.join(new_vehicle_dir, item)
            if os_path.isdir(s):
                shutil_copytree(s, d)
            else:
                shutil_copy2(s, d)

    @staticmethod
    def getcwd():
        return os_getcwd()

    def tempcal_imu_result_param_tuple(self):
        tempcal_imu_result_param_filename = "03_imu_temperature_calibration_results.param"
        return [tempcal_imu_result_param_filename, os_path.join(self.vehicle_dir, tempcal_imu_result_param_filename)]

    def copy_fc_values_to_file(self, selected_file: str, params: Dict[str, float]):
        ret = 0
        if selected_file in self.file_parameters:
            for param, v in self.file_parameters[selected_file].items():
                if param in params:
                    v.value = params[param]
                    ret += 1
                else:
                    logging_warning("Parameter %s not found in the current parameter file", param)
        return ret

    def write_last_uploaded_filename(self, current_file: str):
        try:
            with open(os_path.join(self.vehicle_dir, 'last_uploaded_filename.txt'), 'w', encoding='utf-8') as file:
                file.write(current_file)
        except Exception as e:  # pylint: disable=broad-except
            logging_error("Error writing last uploaded filename: %s", e)

    def __read_last_uploaded_filename(self) -> str:
        try:
            with open(os_path.join(self.vehicle_dir, 'last_uploaded_filename.txt'), 'r', encoding='utf-8') as file:
                return file.read().strip()
        except FileNotFoundError as e:
            logging_debug("last_uploaded_filename.txt not found: %s", e)
        except Exception as e:  # pylint: disable=broad-except
            logging_error("Error reading last uploaded filename: %s", e)
        return ""

    def get_start_file(self, explicit_index: int):
        # Get the list of intermediate parameter files files that will be processed sequentially
        files = list(self.file_parameters.keys())

        if explicit_index >= 0:
            if not files:
                return ""

            # Determine the starting file based on the --n command line argument
            start_file_index = explicit_index # Ensure the index is within the range of available files
            if start_file_index >= len(files):
                start_file_index = len(files) - 1
                logging_warning("Starting file index %s is out of range. Starting with file %s instead.",
                                explicit_index, files[start_file_index])
            return files[start_file_index]

        last_uploaded_filename = self.__read_last_uploaded_filename()
        if last_uploaded_filename:
            logging_info("Last uploaded file was %s.", last_uploaded_filename)
        else:
            logging_info("No last uploaded file found. Starting with the first file.")
            return files[0]

        if last_uploaded_filename not in files:
            # Handle the case where last_uploaded_filename is not found in the list
            logging_warning("Last uploaded file not found in the list of files. Starting with the first file.")
            return files[0]

        # Find the index of last_uploaded_filename in files
        last_uploaded_index = files.index(last_uploaded_filename)
        # Check if there is a file following last_uploaded_filename
        start_file_index = last_uploaded_index + 1
        if start_file_index >= len(files):
            # Handle the case where last_uploaded_filename is the last file in the list
            logging_warning("Last uploaded file is the last file in the list. Starting from there.")
            start_file_index = len(files) - 1
        return files[start_file_index]

    def get_eval_variables(self):
        variables = {}
        if hasattr(self, 'vehicle_components') and self.vehicle_components and \
                'Components' in self.vehicle_components:
            variables['vehicle_components'] = self.vehicle_components['Components']
        if hasattr(self, 'doc_dict') and self.doc_dict:
            variables['doc_dict'] = self.doc_dict
        return variables

    def copy_fc_params_values_to_template_created_vehicle_files(self, fc_parameters: Dict[str, 'Par']):
        eval_variables = self.get_eval_variables()
        for param_filename, param_dict in self.file_parameters.items():
            for param_name, param in param_dict.items():
                if param_name in fc_parameters:
                    param.value = fc_parameters[param_name]
            if self.configuration_steps and param_filename in self.configuration_steps:
                step_dict = self.configuration_steps[param_filename]
                error_msg = self.compute_parameters(param_filename, step_dict, 'forced', eval_variables)
                if error_msg:
                    return error_msg
                error_msg = self.compute_parameters(param_filename, step_dict, 'derived', eval_variables)
                if error_msg:
                    return error_msg
            Par.export_to_param(Par.format_params(param_dict), os_path.join(self.vehicle_dir, param_filename))
        return ''

    @staticmethod
    def supported_vehicles():
        return ['AP_Periph', 'AntennaTracker', 'ArduCopter', 'ArduPlane',
                                    'ArduSub', 'Blimp', 'Heli', 'Rover', 'SITL']

    @staticmethod
    def add_argparse_arguments(parser):
        parser.add_argument('-t', '--vehicle-type',
                            choices=LocalFilesystem.supported_vehicles(),
                            default='',
                            help='The type of the vehicle. Defaults to ArduCopter')
        parser.add_argument('--vehicle-dir',
                            type=str,
                            default=os_getcwd(),
                            help='Directory containing vehicle-specific intermediate parameter files. '
                            'Defaults to the current working directory')
        parser.add_argument('--n',
                            type=int,
                            default=-1,
                            help='Start directly on the nth intermediate parameter file (skips previous files). '
                            'Default is to start on the file next to the last that you wrote to the flight controller.'
                            'If the file does not exist, it will start on the first file.')
        parser.add_argument('--allow-editing-template-files',
                            action='store_true',
                            help='Allow opening and editing template files directly. '
                            'Only for software developers that know what they are doing.')
        return parser



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

import argparse
from logging import basicConfig as logging_basicConfig
from logging import getLevelName as logging_getLevelName

import tkinter as tk
from tkinter import ttk

from MethodicConfigurator.middleware_template_overview import TemplateOverview

from MethodicConfigurator.backend_filesystem_program_settings import ProgramSettings
from MethodicConfigurator.backend_filesystem_vehicle_components import VehicleComponents

from MethodicConfigurator.common_arguments import add_common_arguments_and_parse

from MethodicConfigurator.frontend_tkinter_base import BaseWindow

from MethodicConfigurator.version import VERSION

class TemplateOverviewWindow(BaseWindow):
    """
    Represents the window for viewing and managing ArduPilot vehicle templates.

    This class creates a graphical user interface (GUI) window that displays an overview of available vehicle templates.
    Users can browse through different templates, view their attributes, and perform actions such as storing a template
    directory for further configuration. The window utilizes a Treeview widget to present the templates in a structured
    manner, making it easier for users to navigate and select the desired template for configuration.

    Attributes:
        window (tk.Tk): The root Tkinter window object for the GUI.

    Methods:
        on_row_double_click(event): Handles the event triggered when a row in the Treeview is double-clicked, allowing the user
                                     to store the corresponding template directory.
    """
    def __init__(self, parent: tk.Tk):
        super().__init__(parent)
        self.root.title(f"Amilcar Lucas's - ArduPilot methodic configurator {VERSION} - Template Overview and selection")
        self.root.geometry("1200x300")

        instruction_text = "Please double-click the template below that most resembles your own vehicle components"
        instruction_label = ttk.Label(self.main_frame, text=instruction_text, font=('Arial', 12))
        instruction_label.pack(pady=(10, 20))

        self.sort_column = None

        style = ttk.Style(self.root)
        # Add padding to Treeview heading style
        style.layout("Treeview.Heading", [
            ("Treeview.Heading.cell", {'sticky': 'nswe'}),
            ("Treeview.Heading.border", {'sticky':'nswe', 'children': [
                ("Treeview.Heading.padding", {'sticky':'nswe', 'children': [
                    ("Treeview.Heading.image", {'side':'right', 'sticky':''}),
                    ("Treeview.Heading.text", {'sticky':'we'})
                ]})
            ]}),
        ])
        style.configure("Treeview.Heading", padding=[2, 2, 2, 18], justify='center')

        # Define the columns for the Treeview
        columns = TemplateOverview.columns()
        self.tree = ttk.Treeview(self.main_frame, columns=columns, show='headings')
        for col in columns:
            self.tree.heading(col, text=col)

        # Populate the Treeview with data from the template overview
        for key, template_overview in VehicleComponents.get_vehicle_components_overviews().items():
            attribute_names = template_overview.attributes()
            values = (key,) + tuple(getattr(template_overview, attr, '') for attr in attribute_names)
            self.tree.insert('', 'end', text=key, values=values)

        self.__adjust_treeview_column_widths()

        self.tree.bind('<Double-1>', self.__on_row_double_click)
        self.tree.pack(fill=tk.BOTH, expand=True)

        for col in self.tree["columns"]:
            self.tree.heading(col, text=col, command=lambda col=col: self.__sort_by_column(col, False))

        if isinstance(self.root, tk.Toplevel):
            try:
                while self.root.children:
                    self.root.update_idletasks()
                    self.root.update()
            except tk.TclError as _exp:
                pass
        else:
            self.root.mainloop()

    def __adjust_treeview_column_widths(self):
        """
        Adjusts the column widths of the Treeview to fit the contents of each column.
        """
        for col in self.tree["columns"]:
            max_width = 0
            for subtitle in col.title().split('\n'):
                max_width = max(max_width, tk.font.Font().measure(subtitle))

            # Iterate over all rows and update the max_width if a wider entry is found
            for item in self.tree.get_children():
                item_text = self.tree.item(item, 'values')[self.tree["columns"].index(col)]
                text_width = tk.font.Font().measure(item_text)
                max_width = max(max_width, text_width)

            # Update the column's width property to accommodate the largest text width
            self.tree.column(col, width=int(max_width*0.6 + 10))

    def __on_row_double_click(self, event):
        """Handle row double-click event."""
        item_id = self.tree.identify_row(event.y)
        if item_id:
            selected_template_relative_path = self.tree.item(item_id)['text']
            ProgramSettings.store_template_dir(selected_template_relative_path)
            self.root.destroy()

    def __sort_by_column(self, col, reverse: bool):
        if self.sort_column  and self.sort_column != col:
            self.tree.heading(self.sort_column, text=self.sort_column)
        self.tree.heading(col, text=col + (' ▼' if reverse else ' ▲'))
        self.sort_column = col

        try:
            col_data = [(float(self.tree.set(k, col)), k) for k in self.tree.get_children('')]
        except ValueError:
            col_data = [(self.tree.set(k, col), k) for k in self.tree.get_children('')]
        col_data.sort(reverse=reverse)

        # rearrange items in sorted positions
        for index, (_val, k) in enumerate(col_data):
            self.tree.move(k, '', index)

        # reverse sort next time
        self.tree.heading(col, command=lambda: self.__sort_by_column(col, not reverse))


def argument_parser():
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.
    """
    parser = argparse.ArgumentParser(description='ArduPilot methodic configurator is a GUI-based tool designed to simplify '
                                                 'the management and visualization of ArduPilot parameters. It enables users '
                                                 'to browse through various vehicle templates, edit parameter files, and '
                                                 'apply changes directly to the flight controller. The tool is built to '
                                                 'semi-automate the configuration process of ArduPilot for drones by '
                                                 'providing a clear and intuitive interface for parameter management.')
    return add_common_arguments_and_parse(parser)


def main():
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format='%(asctime)s - %(levelname)s - %(message)s')

    TemplateOverviewWindow(None)

    print(ProgramSettings.get_recently_used_dirs()[0])


if __name__ == "__main__":
    main()



#!/usr/bin/python3

'''
This script updates the PID adjustment parameters to be factor of the corresponding autotuned or optimized parameters.

Usage:
    ./param_pid_adjustment_update.py -d /path/to/directory optimized_parameter_file.param

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

import os
import argparse
import subprocess
from typing import List, Dict
import re

PARAM_NAME_REGEX = r'^[A-Z][A-Z_0-9]*$'
PARAM_NAME_MAX_LEN = 16
VERSION = '1.0'


def parse_arguments():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="""
Updates PID adjustment parameters values based on the given ADJUSTMENT_FACTOR argument.

It loads three sets of parameters from files in the DIRECTORY directory:
    00_default.param - the (complete) default parameters,
    optimized_param_file - the (complete) optimized parameters, and
    16_pid_adjustment.param - the (intermediate) PID adjustment parameters.
It calculates the PID adjustment parameter values based on the ADJUSTMENT_FACTOR argument.
It updates the intermediate parameter file 16_pid_adjustment.param with parameter comments
explaining how their new value relates to the default parameter value.
""")
    parser.add_argument("-d", "--directory",
                        required=True,
                        help="The directory where the parameter files are located.",
                        )
    parser.add_argument("-a", "--adjustment_factor",
                        type=ranged_type(float, 0.1, 0.8), default=0.5,
                        help="The adjustment factor to apply to the optimized parameters. "
                             "Must be in the interval 0.1 to 0.8. Defaults to 0.5.",
                        )
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s {VERSION}',
                        help='Display version information and exit.',
                        )
    parser.add_argument("optimized_param_file",
                        help="The name of the optimized parameter file.",
                        )
    args = parser.parse_args()
    return args


def ranged_type(value_type, min_value, max_value):
    """
    Return function handle of an argument type function for ArgumentParser checking a range:
        min_value <= arg <= max_value
    Args:
        value_type  - value-type to convert arg to
        min_value   - minimum acceptable argument value
        max_value   - maximum acceptable argument value
    """
    def range_checker(arg: str):
        try:
            f = value_type(arg)
        except ValueError as exc:
            raise argparse.ArgumentTypeError(f'must be a valid {value_type}') from exc
        if f < min_value or f > max_value:
            raise argparse.ArgumentTypeError(f'must be within [{min_value}, {max_value}]')
        return f
    # Return function handle to checking function
    return range_checker


class Par:
    """
    A class representing a parameter with a value and an optional comment.

    Attributes:
        value (float): The value of the parameter.
        comment (str): An optional comment describing the parameter.
    """
    def __init__(self, value: float, comment: str = None):
        self.value = value
        self.comment = comment

    @staticmethod
    def load_param_file_into_dict(param_file: str) -> Dict[str, 'Par']:
        parameter_dict = {}
        content = []
        with open(param_file, encoding="utf-8") as f_handle:
            for n, line in enumerate(f_handle, start=1):
                line = line.strip()
                content.append(line)
                comment = None
                if not line or line.startswith("#"):
                    continue
                if "#" in line:
                    line, comment = line.split("#", 1)
                    comment = comment.strip()
                if "," in line:
                    parameter, value = line.split(",", 1)
                elif " " in line:
                    parameter, value = line.split(" ", 1)
                elif "\t" in line:
                    parameter, value = line.split("\t", 1)
                else:
                    raise SystemExit(f"Missing parameter-value separator: {line} in {param_file} line {n}")
                if len(parameter) > PARAM_NAME_MAX_LEN:
                    raise SystemExit(f"Too long parameter name: {parameter} in {param_file} line {n}")
                if not re.match(PARAM_NAME_REGEX, parameter):
                    raise SystemExit(f"Invalid characters in parameter name {parameter} in {param_file} line {n}")
                try:
                    fvalue = float(value)
                except ValueError as exc:
                    raise SystemExit(f"Invalid parameter value {value} in {param_file} line {n}") from exc
                if parameter in parameter_dict:
                    raise SystemExit(f"Duplicated parameter {parameter} in {param_file} line {n}")
                parameter_dict[parameter] = Par(fvalue, comment)
        return parameter_dict, content

    @staticmethod
    def export_to_param(param_dict: Dict[str, 'Par'], filename_out: str, content_header: List[str] = None) -> None:
        if content_header is None:
            content_header = []
        with open(filename_out, "w", encoding="utf-8") as output_file:
            if content_header:
                output_file.write('\n'.join(content_header) + '\n')
            for key, par in param_dict.items():
                line = f"{key},{format(par.value, '.6f').rstrip('0').rstrip('.')}"
                if par.comment:
                    line += f"  # {par.comment}"
                output_file.write(line + "\n")


def update_pid_adjustment_params(directory: str, optimized_param_file: str, adjustment_factor: float) -> None:
    """
    Updates the PID adjustment parameters values based on the given adjustment factor.

    This function loads three sets of parameters from files: the default parameters, the optimized parameters,
    and the PID adjustment parameters.
    It then updates the PID adjustment parameters based on the given adjustment factor.
    Finally, it exports the updated PID adjustment parameters to a file with a comment explaining how their
    new value relates to the default parameter value.

    Args:
        directory (str): The directory where the parameter files are located.
        optimized_param_file (str): The name of the optimized parameter file.
        adjustment_factor (float): The adjustment factor to apply to the optimized parameters.
    """
    default_param_file_path = os.path.join(directory, "00_default.param")
    optimized_param_file_path = os.path.join(directory, optimized_param_file)
    pid_adjustment_file_path = os.path.join(directory, "16_pid_adjustment.param")

    # Load the default parameter file into a dictionary (comment source)
    default_params_dict, _ = Par.load_param_file_into_dict(default_param_file_path)

    # Load the optimized parameter file into a dictionary (source)
    optimized_params_dict, _ = Par.load_param_file_into_dict(optimized_param_file_path)

    # Load the PID adjustment parameter file into a dictionary (destination)
    pid_adjustment_params_dict, content = Par.load_param_file_into_dict(pid_adjustment_file_path)

    if not default_params_dict:
        raise SystemExit(f"Failed to load default parameters from {default_param_file_path}")

    if not optimized_params_dict:
        raise SystemExit(f"Failed to load optimized parameters from {optimized_param_file_path}")

    if not pid_adjustment_params_dict:
        raise SystemExit(f"Failed to load PID adjustment parameters from {pid_adjustment_file_path}")

    # Update the PID adjustment parameters based on the given adjustment factor
    for param_name, param_value in pid_adjustment_params_dict.items():
        if param_name not in optimized_params_dict:
            raise SystemExit(f"Parameter {param_name} is not present in {optimized_param_file_path}")
        if param_name not in default_params_dict:
            raise SystemExit(f"Parameter {param_name} is not present in {default_param_file_path}")
        # adjust the parameter value
        param_value.value = optimized_params_dict[param_name].value * adjustment_factor
        if default_params_dict[param_name].value != 0:
            coef = param_value.value / default_params_dict[param_name].value
        else:
            coef = 1.0
            param_value.value = 0  # if the default is zero, let it stay at zero, it is safer
        # explain how the new value relates to the default parameter value
        param_value.comment = f" = {format(coef, '.6f').rstrip('0').rstrip('.')} * (" \
                              f"{format(default_params_dict[param_name].value, '.6f').rstrip('0').rstrip('.')} default)"

    return pid_adjustment_params_dict, pid_adjustment_file_path, content[0:7]


def main():
    args = parse_arguments()
    # calculate the parameter values and their comments
    pid_adjustment_params_dict, pid_adjustment_file_path, content_header = update_pid_adjustment_params(
        args.directory, args.optimized_param_file, args.adjustment_factor)
    # export the updated PID adjust parameters to a file, preserving the first eight header lines
    Par.export_to_param(pid_adjustment_params_dict, pid_adjustment_file_path, content_header)
    # annotate each parameter with up-to date documentation
    subprocess.run(['./annotate_params.py', os.path.join(args.directory, "16_pid_adjustment.param")], check=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from os import path as os_path

# from sys import exit as sys_exit
# from logging import debug as logging_debug
from logging import info as logging_info
from logging import warning as logging_warning
from logging import error as logging_error

from json import load as json_load
from json import JSONDecodeError

from typing import Tuple

from MethodicConfigurator.annotate_params import Par


class ConfigurationSteps:
    """
    A class to manage configuration steps for the ArduPilot methodic configurator.

    This class provides methods for reading and validating configuration steps, including forced and derived parameters.
    It is designed to simplify the interaction with configuration steps for managing ArduPilot configuration files.

    Attributes:
        configuration_steps_filename (str): The name of the file containing documentation for the configuration files.
        configuration_steps (dict): A dictionary containing the configuration steps.
    """
    def __init__(self, _vehicle_dir: str, vehicle_type: str):
        self.configuration_steps_filename = vehicle_type + "_configuration_steps.json"
        self.configuration_steps = {}
        self.forced_parameters = {}
        self.derived_parameters = {}
        self.log_loaded_file = False

    def re_init(self, vehicle_dir: str, vehicle_type: str):
        self.configuration_steps_filename = vehicle_type + "_configuration_steps.json"
        # Define a list of directories to search for the configuration_steps_filename file
        search_directories = [vehicle_dir, os_path.dirname(os_path.abspath(__file__))]
        file_found = False
        for i, directory in enumerate(search_directories):
            try:
                with open(os_path.join(directory, self.configuration_steps_filename), 'r', encoding='utf-8') as file:
                    self.configuration_steps = json_load(file)
                    file_found = True
                    if self.log_loaded_file:
                        if i == 0:
                            logging_warning("Configuration steps '%s' loaded from %s " \
                                            "(overwriting default configuration steps).",
                                            self.configuration_steps_filename, directory)
                        if i == 1:
                            logging_info("Configuration steps '%s' loaded from %s.",
                                         self.configuration_steps_filename, directory)
                    break
            except FileNotFoundError:
                pass
            except JSONDecodeError as e:
                logging_error("Error in file '%s': %s", self.configuration_steps_filename, e)
                break
        if file_found:
            for filename, file_info in self.configuration_steps.items():
                self.__validate_parameters_in_configuration_steps(filename, file_info, 'forced')
                self.__validate_parameters_in_configuration_steps(filename, file_info, 'derived')
        else:
            logging_warning("No configuration steps documentation and no forced and derived parameters will be available.")
        self.log_loaded_file = True

    def __validate_parameters_in_configuration_steps(self, filename: str, file_info: dict, parameter_type: str) -> None:
        """
        Validates the parameters in the configuration steps.

        This method checks if the parameters in the configuration steps are correctly formatted.
        If a parameter is missing the 'New Value' or 'Change Reason' attribute, an error message is logged.
        """
        if parameter_type + '_parameters' in file_info:
            if not isinstance(file_info[parameter_type + '_parameters'], dict):
                logging_error("Error in file '%s': '%s' %s parameter is not a dictionary",
                             self.configuration_steps_filename, filename, parameter_type)
                return
            for parameter, parameter_info in file_info[parameter_type + '_parameters'].items():
                if "New Value" not in parameter_info:
                    logging_error("Error in file '%s': '%s' %s parameter '%s'"
                                        " 'New Value' attribute not found.",
                                        self.configuration_steps_filename, filename, parameter_type, parameter)
                if "Change Reason" not in parameter_info:
                    logging_error("Error in file '%s': '%s' %s parameter '%s'"
                                        " 'Change Reason' attribute not found.",
                                        self.configuration_steps_filename, filename, parameter_type, parameter)

    def compute_parameters(self, filename: str, file_info: dict, parameter_type: str, variables: dict) -> str:
        """
        Computes the forced or derived parameters for a given configuration file.

        If the parameter is forced, it is added to the forced_parameters dictionary.
        If the parameter is derived, it is added to the derived_parameters dictionary.
        """
        if parameter_type + '_parameters' not in file_info or not variables:
            return ""
        destination = self.forced_parameters if parameter_type == 'forced' else self.derived_parameters
        for parameter, parameter_info in file_info[parameter_type + '_parameters'].items():  # pylint: disable=too-many-nested-blocks
            try:
                if ('fc_parameters' in str(parameter_info["New Value"])) and ('fc_parameters' not in variables):
                    error_msg = f"In file '{self.configuration_steps_filename}': '{filename}' {parameter_type} " \
                        f"parameter '{parameter}' could not be computed: 'fc_parameters' not found, is an FC connected?"
                    if parameter_type == 'forced':
                        logging_error(error_msg)
                        return error_msg
                    logging_warning(error_msg)
                    continue
                result = eval(str(parameter_info["New Value"]), {}, variables)  # pylint: disable=eval-used

                # convert (combobox) string text to (parameter value) string int or float
                if isinstance(result, str):
                    if parameter in variables['doc_dict']:
                        values = variables['doc_dict'][parameter]['values']
                        if values:
                            result = next(key for key, value in values.items() if value == result)
                        else:
                            bitmasks = variables['doc_dict'][parameter]['Bitmask']
                            if bitmasks:
                                result = 2**next(key for key, bitmask in bitmasks.items() if bitmask == result)

                if filename not in destination:
                    destination[filename] = {}
                destination[filename][parameter] = Par(float(result), parameter_info["Change Reason"])
            except (SyntaxError, NameError, KeyError) as e:
                error_msg = f"In file '{self.configuration_steps_filename}': '{filename}' {parameter_type} " \
                    f"parameter '{parameter}' could not be computed: {e}"
                if parameter_type == 'forced':
                    logging_error(error_msg)
                    return error_msg
                logging_warning(error_msg)
        return ""

    def auto_changed_by(self, selected_file: str):
        if selected_file in self.configuration_steps:
            return self.configuration_steps[selected_file].get('auto_changed_by', '')
        return ''

    def jump_possible(self, selected_file: str):
        if selected_file in self.configuration_steps:
            return self.configuration_steps[selected_file].get('jump_possible', {})
        return {}

    def get_documentation_text_and_url(self, selected_file: str, prefix_key: str) -> Tuple[str, str]:
        documentation = self.configuration_steps.get(selected_file, {}) if \
            self.configuration_steps else None
        if documentation is None:
            text = f"File '{self.configuration_steps_filename}' not found. " \
                "No intermediate parameter configuration steps available"
            url = None
        else:
            text = documentation.get(prefix_key + "_text", f"No documentation available for {selected_file} in the "
                                     f"{self.configuration_steps_filename} file")
            url = documentation.get(prefix_key + "_url", None)
        return text, url



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

# adding all this allows pyinstaller to build a working windows executable
# note that using --hidden-import does not work for these modules
try:
    from pymavlink import mavutil
    # import pymavlink.dialects.v20.ardupilotmega
except Exception: # pylint: disable=broad-exception-caught
    pass

class BackendFlightcontrollerInfo:  # pylint: disable=too-many-instance-attributes
    """
    Handle flight controller information.

    It includes methods for setting various attributes such as system ID, component ID,
    autopilot type, vehicle type, and capabilities among others.
    """
    def __init__(self):
        self.system_id = None
        self.component_id = None
        self.autopilot = None
        self.vehicle_type = None
        self.mav_type = None
        self.flight_sw_version = None
        self.flight_sw_version_and_type = None
        self.board_version = None
        self.flight_custom_version = None
        self.os_custom_version = None
        self.vendor = None
        self.vendor_id = None
        self.vendor_and_vendor_id = None
        self.product = None
        self.product_id = None
        self.product_and_product_id = None
        self.capabilities = None

        self.is_supported = False
        self.is_mavftp_supported = False

    def get_info(self):
        return {
            "Vendor": self.vendor_and_vendor_id,
            "Product": self.product_and_product_id,
            "Hardware Version": self.board_version,
            "Autopilot Type": self.autopilot,
            "ArduPilot FW Type": self.vehicle_type,
            "MAV Type": self.mav_type,
            "Firmware Version": self.flight_sw_version_and_type,
            "Git Hash": self.flight_custom_version,
            "OS Git Hash": self.os_custom_version,
            "Capabilities": self.capabilities,
            "System ID": self.system_id,
            "Component ID": self.component_id
        }

    def set_system_id_and_component_id(self, system_id, component_id):
        self.system_id = system_id
        self.component_id = component_id

    def set_autopilot(self, autopilot):
        self.autopilot = self.__decode_mav_autopilot(autopilot)
        self.is_supported = autopilot == mavutil.mavlink.MAV_AUTOPILOT_ARDUPILOTMEGA

    def set_type(self, mav_type):
        self.vehicle_type = self.__classify_vehicle_type(mav_type)
        self.mav_type = self.__decode_mav_type(mav_type)

    def set_flight_sw_version(self, version):
        v_major, v_minor, v_patch, v_fw_type = self.__decode_flight_sw_version(version)
        self.flight_sw_version = f"{v_major}.{v_minor}.{v_patch}"
        self.flight_sw_version_and_type = self.flight_sw_version + " " + v_fw_type

    def set_board_version(self, board_version):
        self.board_version = board_version

    def set_flight_custom_version(self, flight_custom_version):
        self.flight_custom_version = ''.join(chr(c) for c in flight_custom_version)

    def set_os_custom_version(self, os_custom_version):
        self.os_custom_version = ''.join(chr(c) for c in os_custom_version)

    def set_vendor_id_and_product_id(self, vendor_id, product_id):
        pid_vid_dict = self.__list_ardupilot_supported_usb_pid_vid()

        self.vendor_id = f"0x{vendor_id:04X}" if vendor_id else "Unknown"
        if vendor_id and vendor_id in pid_vid_dict:
            self.vendor = f"{pid_vid_dict[vendor_id]['vendor']}"
        elif vendor_id:
            self.vendor = "Unknown"
        self.vendor_and_vendor_id = f"{self.vendor} ({self.vendor_id})"

        self.product_id = f"0x{product_id:04X}" if product_id else "Unknown"
        if vendor_id and product_id and product_id in pid_vid_dict[vendor_id]['PID']:
            self.product = f"{pid_vid_dict[vendor_id]['PID'][product_id]}"
        elif product_id:
            self.product = "Unknown"
        self.product_and_product_id = f"{self.product} ({self.product_id})"

    def set_capabilities(self, capabilities):
        self.capabilities = self.__decode_flight_capabilities(capabilities)
        self.is_mavftp_supported = capabilities & mavutil.mavlink.MAV_PROTOCOL_CAPABILITY_FTP

    @staticmethod
    def __decode_flight_sw_version(flight_sw_version):
        '''decode 32 bit flight_sw_version mavlink parameter
        corresponds to ArduPilot encoding in  GCS_MAVLINK::send_autopilot_version'''
        fw_type_id = (flight_sw_version >>  0) % 256  # noqa E221, E222
        patch      = (flight_sw_version >>  8) % 256  # noqa E221, E222
        minor      = (flight_sw_version >> 16) % 256  # noqa E221
        major      = (flight_sw_version >> 24) % 256  # noqa E221
        if fw_type_id == 0:
            fw_type = "dev"
        elif fw_type_id == 64:
            fw_type = "alpha"
        elif fw_type_id == 128:
            fw_type = "beta"
        elif fw_type_id == 192:
            fw_type = "rc"
        elif fw_type_id == 255:
            fw_type = "official"
        else:
            fw_type = "undefined"
        return major, minor, patch, fw_type


    @staticmethod
    def __decode_flight_capabilities(capabilities):
        '''Decode 32 bit flight controller capabilities bitmask mavlink parameter.
        Returns a dict of concise English descriptions of each active capability.
        '''
        capabilities_dict = {}

        # Iterate through each bit in the capabilities bitmask
        for bit in range(32):
            # Check if the bit is set
            if capabilities & (1 << bit):
                # Use the bit value to get the corresponding capability enum
                capability = mavutil.mavlink.enums["MAV_PROTOCOL_CAPABILITY"].get(1 << bit, "Unknown capability")
                # Append the abbreviated name and description of the capability dictionary
                capabilities_dict[capability.name.replace("MAV_PROTOCOL_CAPABILITY_", "")] = capability.description

        return capabilities_dict


    # see for more info:
    # import pymavlink.dialects.v20.ardupilotmega
    # pymavlink.dialects.v20.ardupilotmega.enums["MAV_TYPE"]
    @staticmethod
    def __decode_mav_type(mav_type):
        return mavutil.mavlink.enums["MAV_TYPE"].get(mav_type,
                                                    mavutil.mavlink.EnumEntry("None", "Unknown type")).description


    @staticmethod
    def __decode_mav_autopilot(mav_autopilot):
        return mavutil.mavlink.enums["MAV_AUTOPILOT"].get(mav_autopilot,
                                                        mavutil.mavlink.EnumEntry("None", "Unknown type")).description


    @staticmethod
    def __classify_vehicle_type(mav_type_int):
        """
        Classify the vehicle type based on the MAV_TYPE enum.

        Parameters:
        mav_type_int (int): The MAV_TYPE enum value.

        Returns:
        str: The classified vehicle type.
        """
        # Define the mapping from MAV_TYPE_* integer to vehicle type category
        mav_type_to_vehicle_type = {
            mavutil.mavlink.MAV_TYPE_FIXED_WING: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_QUADROTOR: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_COAXIAL: 'Heli',
            mavutil.mavlink.MAV_TYPE_HELICOPTER: 'Heli',
            mavutil.mavlink.MAV_TYPE_ANTENNA_TRACKER: 'AntennaTracker',
            mavutil.mavlink.MAV_TYPE_GCS: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_AIRSHIP: 'ArduBlimp',
            mavutil.mavlink.MAV_TYPE_FREE_BALLOON: 'ArduBlimp',
            mavutil.mavlink.MAV_TYPE_ROCKET: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_GROUND_ROVER: 'Rover',
            mavutil.mavlink.MAV_TYPE_SURFACE_BOAT: 'Rover',
            mavutil.mavlink.MAV_TYPE_SUBMARINE: 'ArduSub',
            mavutil.mavlink.MAV_TYPE_HEXAROTOR: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_OCTOROTOR: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_TRICOPTER: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_FLAPPING_WING: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_KITE: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_ONBOARD_CONTROLLER: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_VTOL_DUOROTOR: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_VTOL_QUADROTOR: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_VTOL_TILTROTOR: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_VTOL_RESERVED2: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_VTOL_RESERVED3: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_VTOL_RESERVED4: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_VTOL_RESERVED5: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_GIMBAL: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_ADSB: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_PARAFOIL: 'ArduPlane',
            mavutil.mavlink.MAV_TYPE_DODECAROTOR: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_CAMERA: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_CHARGING_STATION: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_FLARM: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_SERVO: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_ODID: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_DECAROTOR: 'ArduCopter',
            mavutil.mavlink.MAV_TYPE_BATTERY: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_PARACHUTE: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_LOG: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_OSD: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_IMU: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_GPS: 'AP_Periph',
            mavutil.mavlink.MAV_TYPE_WINCH: 'AP_Periph',
            # Add more mappings as needed
        }

        # Return the classified vehicle type based on the MAV_TYPE enum
        return mav_type_to_vehicle_type.get(mav_type_int, None)

    @staticmethod
    def __list_ardupilot_supported_usb_pid_vid():
        """
        List all ArduPilot supported USB vendor ID (VID) and product ID (PID).

        source: https://ardupilot.org/dev/docs/USB-IDs.html
        """
        return {
            0x0483: {'vendor': 'ST Microelectronics', 'PID': {0x5740: 'ChibiOS'}},
            0x1209: {'vendor': 'ArduPilot', 'PID': {0x5740: 'MAVLink',
                                                    0x5741: 'Bootloader',
                                                    }
                     },
            0x16D0: {'vendor': 'ArduPilot', 'PID': {0x0E65: 'MAVLink'}},
            0x26AC: {'vendor': '3D Robotics', 'PID': {}},
            0x2DAE: {'vendor': 'CubePilot', 'PID': {0x1101: 'CubeBlack+',
                                                    0x1001: 'CubeBlack bootloader',
                                                    0x1011: 'CubeBlack',
                                                    0x1002: 'CubeYellow bootloader',
                                                    0x1012: 'CubeYellow',
                                                    0x1004: 'CubeGreen bootloader',
                                                    0x1014: 'CubeGreen',
                                                    0x1005: 'CubePurple bootloader',
                                                    0x1015: 'CubePurple',
                                                    0x1006: 'CubeOrange bootloader',
                                                    0x1016: 'CubeOrange',
                                                    0x1003: 'CubeBlue bootloader',
                                                    0x1013: 'CubeBlue',
                                                    0x1007: 'CubeRed bootloader',
                                                    0x1017: 'CubeRed'
                                                    }
                     },
            0x3162: {'vendor': 'Holybro', 'PID': {0x004B: 'Durandal'}},
            0x27AC: {'vendor': 'Laser Navigation', 'PID': {0x1151: 'VRBrain-v51',
                                                           0x1152: 'VRBrain-v52',
                                                           0x1154: 'VRBrain-v54',
                                                           0x1910: 'VRCore-v10',
                                                           0x1351: 'VRUBrain-v51',
                                                           }
                     },
        }



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

from sys import exit as sys_exit

from argparse import ArgumentParser

from logging import basicConfig as logging_basicConfig
from logging import getLevelName as logging_getLevelName
from logging import debug as logging_debug
from logging import warning as logging_warning
from logging import critical as logging_critical

import tkinter as tk
from tkinter import ttk
from tkinter import simpledialog

from MethodicConfigurator.common_arguments import add_common_arguments_and_parse

from MethodicConfigurator.backend_flightcontroller import FlightController

from MethodicConfigurator.frontend_tkinter_base import show_no_connection_error
from MethodicConfigurator.frontend_tkinter_base import show_tooltip
from MethodicConfigurator.frontend_tkinter_base import update_combobox_width
from MethodicConfigurator.frontend_tkinter_base import ProgressWindow
from MethodicConfigurator.frontend_tkinter_base import BaseWindow


# https://dev.to/geraldew/python-tkinter-an-exercise-in-wrapping-the-combobox-ndb
class PairTupleCombobox(ttk.Combobox):  # pylint: disable=too-many-ancestors
    """
    A custom Combobox widget that allows for the display of a list of tuples, where each tuple contains a key and a value.
    This widget processes the list of tuples to separate keys and values for display purposes and allows for the selection
    of a tuple based on its key.
    """
    def process_list_pair_tuple(self, list_pair_tuple):
        r_list_keys = []
        r_list_shows = []
        if isinstance(list_pair_tuple, list):
            for tpl in list_pair_tuple:
                r_list_keys.append(tpl[0])
                r_list_shows.append(tpl[1])
        elif isinstance(list_pair_tuple, dict):
            for key, value in list_pair_tuple.items():
                r_list_keys.append(key)
                r_list_shows.append(value)
        else:
            logging_critical("list_pair_tuple must be a tuple or a dictionary, not %s", type(list_pair_tuple))
            sys_exit(1)
        return r_list_keys, r_list_shows

    def __init__(self, container, list_pair_tuple, selected_element, cb_name, *args, **kwargs):
        super().__init__(container, *args, **kwargs)
        self.cb_name = cb_name
        self.set_entries_tupple(list_pair_tuple, selected_element)

    def set_entries_tupple(self, list_pair_tuple, selected_element):
        self.list_keys, self.list_shows = self.process_list_pair_tuple(list_pair_tuple)
        self['values'] = tuple(self.list_shows)
        # still need to set the default value from the nominated key
        if selected_element:
            try:
                default_key_index = self.list_keys.index(selected_element)
                self.current(default_key_index)
            except IndexError:
                logging_critical("%s combobox selected string '%s' not in list %s",
                                 self.cb_name, selected_element, self.list_keys)
                sys_exit(1)
            except ValueError:
                logging_critical("%s combobox selected string '%s' not in list %s",
                                 self.cb_name, selected_element, self.list_keys)
                sys_exit(1)
            update_combobox_width(self)
        else:
            logging_warning("No %s combobox element selected", self.cb_name)

    def get_selected_key(self):
        try:
            i_index = self.current()
            return self.list_keys[i_index]
        except IndexError:
            return None


class ConnectionSelectionWidgets():  # pylint: disable=too-many-instance-attributes
    """
    A class for managing the selection of flight controller connections in the GUI.

    This class provides functionality for displaying available flight controller connections,
    allowing the user to select a connection, and handling the connection process.
    """
    def __init__(self, parent, parent_frame, flight_controller: FlightController,  # pylint: disable=too-many-arguments
                 destroy_parent_on_connect: bool, download_params_on_connect: bool):
        self.parent = parent
        self.flight_controller = flight_controller
        self.destroy_parent_on_connect = destroy_parent_on_connect
        self.download_params_on_connect = download_params_on_connect
        self.previous_selection = flight_controller.comport.device if hasattr(self.flight_controller.comport, "device") \
            else None
        self.connection_progress_window = None

        # Create a new frame for the flight controller connection selection label and combobox
        self.container_frame = ttk.Frame(parent_frame)

        # Create a description label for the flight controller connection selection
        conn_selection_label = ttk.Label(self.container_frame, text="flight controller connection:")
        conn_selection_label.pack(side=tk.TOP) # Add the label to the top of the conn_selection_frame

        # Create a read-only combobox for flight controller connection selection
        self.conn_selection_combobox = PairTupleCombobox(self.container_frame, self.flight_controller.get_connection_tuples(),
                                                         self.previous_selection,
                                                        "FC connection",
                                                         state='readonly')
        self.conn_selection_combobox.bind("<<ComboboxSelected>>", self.on_select_connection_combobox_change)
        self.conn_selection_combobox.pack(side=tk.TOP, pady=(4, 0))
        show_tooltip(self.conn_selection_combobox, "Select the flight controller connection\nYou can add a custom connection "
                     "to the existing ones")

    def on_select_connection_combobox_change(self, _event):
        selected_connection = self.conn_selection_combobox.get_selected_key()
        logging_debug(f"Connection combobox changed to: {selected_connection}")
        if self.flight_controller.master is None or selected_connection != self.flight_controller.comport.device:
            if selected_connection == 'Add another':
                if not self.add_connection() and self.previous_selection:
                    # nothing got selected revert to the current connection
                    self.conn_selection_combobox.set(self.previous_selection)
                return
            self.reconnect(selected_connection)

    def add_connection(self):
        # Open the connection selection dialog
        selected_connection = simpledialog.askstring("Flight Controller Connection",
                                                     "Enter the connection string to the flight controller. "
                                                     "Examples are:\n\nCOM4 (on windows)\n"
                                                     "/dev/serial/by-id/usb-xxx (on linux)\n"
                                                     "tcp:127.0.0.1:5761\n"
                                                     "udp:127.0.0.1:14551")
        if selected_connection:
            logging_debug(f"Will add new connection: {selected_connection} if not duplicated")
            self.flight_controller.add_connection(selected_connection)
            connection_tuples = self.flight_controller.get_connection_tuples()
            logging_debug(f"Updated connection tuples: {connection_tuples} with selected connection: {selected_connection}")
            self.conn_selection_combobox.set_entries_tupple(connection_tuples, selected_connection)
            self.reconnect(selected_connection)
        else:
            logging_debug(f"Add connection canceled or string empty {selected_connection}")
        return selected_connection

    def reconnect(self, selected_connection: str = ""):  # defaults to auto-connect
        self.connection_progress_window = ProgressWindow(self.parent.root, "Connecting with the FC",
                                                         "Connection step {} of {}")
        error_message = self.flight_controller.connect(selected_connection,
                                                       self.connection_progress_window.update_progress_bar)
        if error_message:
            show_no_connection_error(error_message)
            return True
        self.connection_progress_window.destroy()
        # Store the current connection as the previous selection
        self.previous_selection = self.flight_controller.comport.device
        if self.destroy_parent_on_connect:
            self.parent.root.destroy()
        if self.download_params_on_connect and hasattr(self.parent, "download_flight_controller_parameters"):
            self.parent.download_flight_controller_parameters(redownload=False)
        return False


class ConnectionSelectionWindow(BaseWindow):
    """
    A window for selecting a flight controller connection.

    This class provides a graphical user interface for selecting a connection to a flight controller.
    It inherits from the BaseWindow class and uses the ConnectionSelectionWidgets class to handle
    the UI elements related to connection selection.
    """
    def __init__(self, flight_controller: FlightController, connection_result_string: str):
        super().__init__()
        self.root.title("Flight controller connection")
        self.root.geometry("460x450") # Set the window size

        # Explain why we are here
        if flight_controller.comport is None:
            introduction_text = "No ArduPilot flight controller was auto-detected detected yet."
        else:
            if ":" in connection_result_string:
                introduction_text = connection_result_string.replace(":", ":\n")
            else:
                introduction_text = connection_result_string
        self.introduction_label = ttk.Label(self.main_frame, anchor=tk.CENTER, justify=tk.CENTER,
                                            text=introduction_text + "\nChoose one of the following three options:")
        self.introduction_label.pack(expand=False, fill=tk.X, padx=6, pady=6)

        # Option 1 - Auto-connect
        option1_label = ttk.Label(text="Auto connection", style="Bold.TLabel")
        option1_label_frame = ttk.LabelFrame(self.main_frame, labelwidget=option1_label, borderwidth=2, relief="solid")
        option1_label_frame.pack(expand=False, fill=tk.X, padx=6, pady=6)
        option1_label = ttk.Label(option1_label_frame, anchor=tk.CENTER, justify=tk.CENTER,
                                  text="Connect a flight controller to the PC,\n"
                                  "wait 7 seconds for it to fully boot and\n"
                                  "press the Auto-connect button below to connect to it")
        option1_label.pack(expand=False, fill=tk.X, padx=6)
        autoconnect_button = ttk.Button(option1_label_frame, text="Auto-connect", command=self.fc_autoconnect)
        autoconnect_button.pack(expand=False, fill=tk.X, padx=100, pady=6)
        show_tooltip(autoconnect_button, "Auto-connect to a 'Mavlink'-talking serial device")

        # Option 2 - Manually select the flight controller connection or add a new one
        option2_label = ttk.Label(text="Manual connection", style="Bold.TLabel")
        option2_label_frame = ttk.LabelFrame(self.main_frame, labelwidget=option2_label, borderwidth=2, relief="solid")
        option2_label_frame.pack(expand=False, fill=tk.X, padx=6, pady=6)
        option2_label = ttk.Label(option2_label_frame, anchor=tk.CENTER, justify=tk.CENTER,
                                  text="Connect a flight controller to the PC,\n"
                                  "wait 7 seconds for it to fully boot and\n"
                                  "manually select the fight controller connection or add a new one")
        option2_label.pack(expand=False, fill=tk.X, padx=6)
        self.connection_selection_widgets = ConnectionSelectionWidgets(self, option2_label_frame, flight_controller,
                                                                       destroy_parent_on_connect=True,
                                                                       download_params_on_connect=False)
        self.connection_selection_widgets.container_frame.pack(expand=False, fill=tk.X, padx=80, pady=6)

        # Option 3 - Skip FC connection, just edit the .param files on disk
        option3_label = ttk.Label(text="No connection", style="Bold.TLabel")
        option3_label_frame = ttk.LabelFrame(self.main_frame, labelwidget=option3_label, borderwidth=2, relief="solid")
        option3_label_frame.pack(expand=False, fill=tk.X, padx=6, pady=6)
        #option3_label = ttk.Label(option3_label_frame, anchor=tk.CENTER, justify=tk.CENTER,
        #                          text="Skip the flight controller connection,\n"
        #                          "no default parameter values will be fetched from the FC,\n"
        #                          "default parameter values from disk will be used instead\n"
        #                          "(if '00_default.param' file is present)\n"
        #                          "and just edit the intermediate '.param' files on disk")
        #option3_label.pack(expand=False, fill=tk.X, padx=6)
        skip_fc_connection_button = ttk.Button(option3_label_frame,
                                              text="Skip FC connection, just edit the .param files on disk",
                                              command=lambda flight_controller=flight_controller:
                                              self.skip_fc_connection(flight_controller))
        skip_fc_connection_button.pack(expand=False, fill=tk.X, padx=15, pady=6)
        show_tooltip(skip_fc_connection_button,
                     "No parameter values will be fetched from the FC, default parameter values from disk will be used\n"
                     "instead (if '00_default.param' file is present) and just edit the intermediate '.param' files on disk")

        # Bind the close_connection_and_quit function to the window close event
        self.root.protocol("WM_DELETE_WINDOW", self.close_and_quit)

    def close_and_quit(self):
        sys_exit(0)

    def fc_autoconnect(self):
        self.connection_selection_widgets.reconnect()

    def skip_fc_connection(self, flight_controller: FlightController):
        logging_warning("Will proceed without FC connection. FC parameters will not be downloaded nor uploaded")
        logging_warning("Only the intermediate '.param' files on the PC disk will be edited")
        flight_controller.disconnect()
        self.root.destroy()


def argument_parser():
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.
    """
    parser = ArgumentParser(description='This main is for testing and development only. '
                            'Usually, the ConnectionSelectionWidgets is called from another script')
    parser = FlightController.add_argparse_arguments(parser)
    return add_common_arguments_and_parse(parser)


def main():
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format='%(asctime)s - %(levelname)s - %(message)s')

    logging_warning("This main is for testing and development only, usually the ConnectionSelectionWindow is called from "
                    "another script")

    flight_controller = FlightController(args.reboot_time) # Initialize your FlightController instance
    result = flight_controller.connect(device=args.device) # Connect to the flight controller
    if result:
        logging_warning(result)
        window = ConnectionSelectionWindow(flight_controller, result)
        window.root.mainloop()
    flight_controller.disconnect() # Disconnect from the flight controller


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

'''
This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
'''

#from logging import debug as logging_debug
from logging import info as logging_info

import tkinter as tk
from tkinter import ttk

from MethodicConfigurator.backend_flightcontroller import FlightController
#from MethodicConfigurator.backend_flightcontroller_info import BackendFlightcontrollerInfo

#from MethodicConfigurator.frontend_tkinter_base import show_tooltip
from MethodicConfigurator.frontend_tkinter_base import ProgressWindow
from MethodicConfigurator.frontend_tkinter_base import BaseWindow

from MethodicConfigurator.version import VERSION


class FlightControllerInfoWindow(BaseWindow):
    """
    Display flight controller hardware, firmware and parameter information
    """
    def __init__(self, flight_controller: FlightController):
        super().__init__()
        self.root.title("ArduPilot methodic configurator " + VERSION + " - Flight Controller Info")
        self.root.geometry("500x350")  # Adjust the window size as needed
        self.flight_controller = flight_controller

        # Create a frame to hold all the labels and text fields
        self.info_frame = ttk.Frame(self.main_frame)
        self.info_frame.pack(padx=20, pady=20)

        # Dynamically create labels and text fields for each attribute
        for row_nr, (description, attr_value) in enumerate(flight_controller.info.get_info().items()):
            label = ttk.Label(self.info_frame, text=f"{description}:")
            label.grid(row=row_nr, column=0, sticky="w")

            text_field = ttk.Entry(self.info_frame, width=60)
            text_field.grid(row=row_nr, column=1, sticky="w")

            # Check if the attribute exists and has a non-empty value before inserting
            if attr_value:
                if isinstance(attr_value, dict):
                    text_field.insert(tk.END, (", ").join(attr_value.keys()))
                else:
                    text_field.insert(tk.END, attr_value)
            else:
                text_field.insert(tk.END, "N/A")  # Insert "Not Available" if the attribute is missing or empty
            text_field.configure(state="readonly")

        logging_info("Firmware Version: %s", flight_controller.info.flight_sw_version_and_type)
        logging_info(f"Firmware first 8 hex bytes of the FC git hash: {flight_controller.info.flight_custom_version}")
        logging_info(f"Firmware first 8 hex bytes of the ChibiOS git hash: {flight_controller.info.os_custom_version}")
        logging_info(f"Flight Controller HW / board version: {flight_controller.info.board_version}")
        logging_info(f"Flight Controller USB vendor ID: {flight_controller.info.vendor}")
        logging_info(f"Flight Controller USB product ID: {flight_controller.info.product}")

        self.root.after(50, self.download_flight_controller_parameters()) # 50 milliseconds
        self.root.mainloop()

    def download_flight_controller_parameters(self):
        param_download_progress_window = ProgressWindow(self.root, "Downloading FC parameters",
                                                        "Downloaded {} of {} parameters")
        self.flight_controller.fc_parameters = self.flight_controller.download_params(
            param_download_progress_window.update_progress_bar)
        param_download_progress_window.destroy()  # for the case that '--device test' and there is no real FC connected
        self.root.destroy()
